(** Runtime evaluation of queries *)

open Str
open Num
open List

open Utility
open Result
open Syntax
open Expression
open Expression.Expr
open Expression.Query

open Unix

exception RuntimeUndefVar of string
exception NoSuchField of string

(* Lookup name in locals, globals, and primitives *)
(* This is used by both Doquery and Interpreter *)
let lookup globals locals name = 
  try 
    (match Utility.lookup name locals with
      | Some v -> v
      | None -> match Utility.lookup name globals with
          | Some v -> v
          | None -> Library.primitive_stub name)
  with NotFound _ -> 
    (* may be thrown by primitive_stub *)
    raise(RuntimeUndefVar name)

(** Substitutes values for the variables in a query,
    and performs interpolation in LIKE expressions. *)
let rec normalise_query (globals:environment) (env:environment) (db:database) (qry:query) : query =
  let rec normalise_like_expression (l : Query.like_expr): Query.expression = 
    Text (Sql_transform.like_as_string (env @ globals) l)
  in
  let rec normalise_expression : Query.expression -> Query.expression = function
      | Query.Variable name ->
          (try
             match lookup globals env name with
               | `Bool value -> Query.Boolean value
               | `Int value -> Query.Integer value
               | `Float value -> Query.Float value
               | `List (`Char _::_) as c  
                 -> Query.Text (db # escape_string (charlist_as_string c))
               | `List ([]) -> Query.Text ""
               | r -> Query.Text (db # escape_string (Result.string_of_result r))
(*                | r -> failwith("Internal error: variable " ^ name ^  *)
(*                                  " in query "^ Sql.string_of_query qry ^  *)
(*                                  " had unexpected type at runtime: " ^  *)
(*                                  string_of_result r) *)
           with NotFound _-> failwith("Internal error: undefined query variable '"
                                     ^ name ^ "'"))
      | Query.Binary_op (symbol, left, right) ->
          Binary_op (symbol, normalise_expression left, normalise_expression right)
      | Query.Unary_op (symbol, expr) ->
          Unary_op (symbol, normalise_expression expr)
      | Query.LikeExpr (like_expr) -> normalise_like_expression like_expr
      | Query.Query qry ->
          Query {qry with condition = normalise_expression qry.condition}
      | Query.Funcall(fname, args) -> 
	  Query.Funcall(fname, List.map normalise_expression args)
      | Query.Case branches -> 
	  Query.Case (List.map (fun (g,t) -> (normalise_expression g, normalise_expression t)) branches)
      | expr -> expr
  in
  let normalise_tables = 
    map (function 
           | `TableName t, alias -> `TableName t, alias
           | `TableVariable var, alias ->
               (match lookup globals env var with
                    `Table(_, tableName, _, _) -> `TableName tableName, alias
                  | _ -> failwith ("Internal Error: "^
                      "table source was not a table!"))
        ) 
  in
  let normalise_cols cols = 
    List.map (fun lr -> match lr with 
		  Left _ -> lr
		| Right ce -> Right {ce with expression = normalise_expression ce.expression}) cols 
  in
    {qry with
       result_cols = normalise_cols qry.result_cols;
       tables = normalise_tables qry.tables;
       condition = normalise_expression qry.condition;
       offset = normalise_expression qry.offset;
       max_rows = opt_map normalise_expression qry.max_rows}

(** Run a query
    row_field_present should really be defined in this module, 
    but it is too dependent on Interpreter *)
let do_query globals locals row_field_present
    (query : Syntax.Sql.Query.query)
    (table_aliases : string list)
    (tables : Result.result list)
    : Result.result =
(*   let row_field_present field fields tc t = *)
(* 	try *)
(* 	  let module E = Expression.Expr in *)
(* 	  let marsh = List.assoc field tc.E.marshallby in *)
(* 		match marsh with *)
(* 		    E.Record_intro(fields, _, _) -> *)
(* 			  let unmarsh = *)
(* 			    (try *)
(* 			       StringMap.find "2" fields *)
(* 			     with *)
(* 			         NotFound _ -> raise *)
(*                        (Runtime_error ("Expected a tuple in table "^ *)
(*                                          "field marshalling constraint"))) in *)
(* 			  let unmarshall s = *)
(* 			    let module E = Expression.Expr in *)
(* 			    let s' = E.Constant (Expression.String s, no_expr_data) in *)
(* 			    let app = E.Apply (unmarsh, [s'], no_expr_data) in *)
(* 			      try *)
(* 			        interpret globals locals app Result.toplevel_cont *)
(* 			      with *)
(* 				      TopLevel (_,v) -> v in *)
(* 			    (t, Some unmarshall) *)
(* 		  | _ -> raise *)
(*               (Runtime_error *)
(*                  ("Expected a tuple in table field marshalling constraint")) *)
(* 	with *)
(* 		Not_found -> (t, None) *)
(*   in *)

  (* lookup field in table definition *)
  let row_field_type field : (Types.row * Syntax.table_constr) -> (Types.datatype * (string -> Result.result) option) =
    fun ((fields,_), tc) ->
      match StringMap.find field fields with
	    | `Present t -> row_field_present globals locals field fields tc t
	    | `Absent -> raise (NoSuchField field) in

  let query_result_types (query : query) (table_defs : (string * (Types.row * Syntax.table_constr)) list)
      : (string * (Types.datatype * (string -> Result.result) option)) list =
    try
	  concat_map
        (function
           | Left col -> [col.col_alias,
                          row_field_type col.name (assoc col.table_alias table_defs)]
           | Right _ -> [])
        query.Query.result_cols
    with NoSuchField field ->
      failwith ("Field " ^ field ^ " from " ^ Sql.string_of_query query
                ^ " was not found in tables " ^ mapstrcat "," fst table_defs ^ ".") in

  (* translate table into table definitions and a single database *)
  let db, table_defs =
    let (dbs, table_defs) =
      split (map
               (function
                    `Table((db, _), _table_name, row, tc) -> (db, (row,tc))
                  | _ -> assert false)
               tables) in
    let db =
      if (all_equiv (=) dbs) then hd(dbs)
      else failwith ("Cannot join across different databases") in
    let table_defs = combine table_aliases table_defs in
      (db, table_defs) in
    
  let result_types = query_result_types query table_defs in
  let query_string = 
    Sql.string_of_query (normalise_query globals locals db query) in
  Debug.f "RUNNING QUERY:\n%s" query_string;
  let t = Unix.gettimeofday() in
  let result = Database.execute_select result_types query_string db in
  Debug.f "Query took : %fs" (Unix.gettimeofday() -. t);
  result
