(*pp deriving *)
open List

open Utility
open Syntax
open Expression
open Expression.Expr
open Expression.Types
open Expression.FableType
open Expression.DerivingSyntax
open Forms
open Errors
open Instantiate
open Typevarcheck
open Typeenv
open Printsyntax

(* debug flags *)
let show_unification = Settings.add_bool("show_unification", false, `User)
let show_row_unification = Settings.add_bool("show_row_unification", false, `User)

let show_instantiation = Settings.add_bool("show_instantiation", false, `User)
let show_generalisation = Settings.add_bool("show_generalisation", false, `User)

let show_typechecking = Settings.add_bool("show_typechecking", false, `User)
let show_recursion = Instantiate.show_recursion

let rigid_type_variables = Settings.add_bool("rigid_type_variables", true, `User)

(*
  what kind of recursive types to allow
  "all"      - allow all recursive types
  "guarded"  - only allow guarded recursive types
  "positive" - only allow positive recursive types
*)
let infer_recursive_types = Settings.add_string("infer_recursive_types", "guarded", `User)

module Unify : sig 
  val unify : string list -> Typeenv.omega -> expression option -> bool -> (Types.datatype * Types.datatype) -> Subst.subst
  val unify_old : Typeenv.omega -> (Types.datatype * Types.datatype) -> unit
  val unify_rows : Typeenv.omega -> bool ->  (Types.row * Types.row) -> Subst.subst
  exception Unify_failure of string
end = struct
  exception Unify_failure of string

  let occurs_check alias_env var t =
    match Settings.get_value infer_recursive_types with
      | "all" -> true
      | "guarded" -> is_guarded alias_env var t
      | "positive" -> not (is_negative alias_env var t)
      | s -> failwith ("user setting infer_recursive_types ("^ s ^") must be set to 'all', 'guarded' or 'positive'")
          
  let occurs_check_row alias_env var row =
    match Settings.get_value infer_recursive_types with
      | "all" -> true
      | "guarded" -> is_guarded_row alias_env var row
      | "positive" -> not (is_negative_row alias_env var row)
      | s -> failwith ("user setting infer_recursive_types ("^ s ^") must be set to 'all', 'guarded' or 'positive'")

  let var_is_free_in_type var datatype = TypeVarSet.mem var (free_type_vars datatype)

  (* a special kind of structural equality on types that doesn't look
     inside points *)
  let rec eq_types : (datatype * datatype) -> bool =
    fun (t1, t2) ->
      match (t1.Types.ltype, t2.Types.ltype) with
        | `Not_typed, `Not_typed -> true
        | `Primitive x, `Primitive y -> x = y
        | `MetaTypeVar lpoint, `MetaTypeVar rpoint ->
            Unionfind.equivalent lpoint rpoint
        | `Function (lfrom, lm, lto), `Function (rfrom, rm, rto) ->
            eq_types (lfrom, rfrom) && eq_types (lto, rto) && eq_types (lm, rm)
        | `Record l, `Record r -> eq_rows (l, r)
        | `Variant l, `Variant r -> eq_rows (l, r)
        | `Application (s, ts), `Application (s', ts') when s = s' -> List.for_all2 (Utility.curry eq_types) ts ts'
        | _, _ -> false
  and eq_rows : (row * row) -> bool =
    fun ((lfield_env, lrow_var), (rfield_env, rrow_var)) ->
      eq_field_envs (lfield_env, rfield_env) && eq_row_vars (lrow_var, rrow_var)
  and eq_field_envs (lfield_env, rfield_env) =
    let compare_specs = fun a b -> 
      match (a,b) with
        | `Absent, `Absent -> true
        | `Present t1, `Present t2 -> eq_types (t1, t2)
        | _, _ -> false
    in
      StringMap.equal compare_specs lfield_env rfield_env
  and eq_row_vars (lpoint, rpoint) =
    (* [QUESTION]
       Do we need to deal with closed rows specially?
    *)
    match Unionfind.find lpoint, Unionfind.find rpoint with
      | ClosedRow, ClosedRow -> true
      | FlexibleRow var, FlexibleRow var'
      | RigidRow var, RigidRow var'
      | RecursiveRow (var, _), RecursiveRow (var', _) -> var=var'
      | _, _ -> Unionfind.equivalent lpoint rpoint

  (*
    unification environment:
    for stopping cycles during unification
    and for type aliases
  *)
  type unify_type_env = (datatype list) IntMap.t
  type unify_row_env = (row list) IntMap.t
  type unify_env = {rec_types : unify_type_env;
                    rec_rows : unify_row_env;
                    omega : Typeenv.omega;
                    phantoms : string list;
		            exp : expression option;
		            polarity : [`Pos | `Neg];
		            subtype : bool}
      
  let add_rec_types var tl uenv = {uenv with rec_types=IntMap.add var tl uenv.rec_types}
  let add_rec_rows var rl uenv = {uenv with rec_rows=IntMap.add var rl uenv.rec_rows}
  let contains_phantom n uenv = List.mem n uenv.phantoms
  let flip_polarity env = {env with polarity = (match env.polarity with `Pos -> `Neg | `Neg -> `Pos)}
  let set_expression env e = {env with exp = Some e}
  let clear_expression env  = {env with exp = None}
    
  let rec unify' : unify_env -> (datatype * datatype) -> Subst.subst = 
    (*     let rec_types, rec_rows, alias_env = rec_env in *)
    let check_polarity uenv = assert_msg (uenv.polarity = `Pos) "unification polarity failure" in

    let is_unguarded_recursive t =
      let rec is_unguarded rec_types t = 
        match t with
          | `MetaTypeVar point ->
              begin
                match Unionfind.find point with
                  | Recursive (var, body) when IntSet.mem var rec_types -> true
                  | Recursive (var, body) -> is_unguarded (IntSet.add var rec_types) (body.Types.ltype)
                  | Body t -> is_unguarded rec_types (t.Types.ltype)
                  | _ -> false
              end
          |  _ -> false
      in
        is_unguarded IntSet.empty t in
      
    let unify_rec uenv ((var, body), t) =
      (*       check_polarity uenv; *)
      let ts =
        if IntMap.mem var uenv.rec_types then
          IntMap.find var uenv.rec_types
        else
          [body]
      in
        (* break cycles *)
        if List.exists (fun t' -> eq_types (t, t')) ts then
          Subst.empty
        else
          unify' (add_rec_types var (t::ts) uenv) (body, t) in
      
    let unify_rec2 uenv ((lvar, lbody), (rvar, rbody)) =
      (*       check_polarity uenv; *)
      let lts =
        if IntMap.mem lvar uenv.rec_types then
          IntMap.find lvar uenv.rec_types
        else
          [lbody] in
        
      let rts =
        if IntMap.mem rvar uenv.rec_types then
          IntMap.find rvar uenv.rec_types
        else
          [rbody]
      in
        (* break cycles *)
        if (List.exists (fun t -> eq_types (t, rbody)) lts
            || List.exists (fun t -> eq_types (t, lbody)) rts) then
          Subst.empty
        else
          unify' ((add_rec_types lvar (rbody::lts) ->- add_rec_types rvar (lbody::rts)) uenv) (lbody, rbody) in
      

    (* introduce a recursive type
       give an error if it is non-well-founded and
       non-well-founded type inference is switched off

       preconditions:
       - Unionfind.find point = t
       - var is free in t
    *)
    let rec_intro uenv point (var, t) =
      (*       check_polarity uenv; *)
      if occurs_check (Typeenv.alias_env uenv.omega) var t then
        Unionfind.change point (Recursive (var, t))
      else
        raise (Unify_failure ("Cannot unify type variable "^string_of_int var^" with datatype "^string_of_datatype t^
                                " because "^
                                match Settings.get_value infer_recursive_types with
                                  | "guarded" -> "the type variable occurs unguarded inside the datatype"
                                  | "positive" -> "the type variable occurs in a negative position inside the datatype"
                                  | _ -> assert false)) in
      
    let lookup_alias (s, ts) uenv =
      try 
	    let (q,d) = Typeenv.lookup_alias (s, ts) uenv.omega in
	      (q, Subst.rename_bound_vars d)
      with
          AliasMismatch msg -> raise (Unify_failure msg) in

    let unify_mtv uenv (mtv : datatype) (t:datatype) : Subst.subst = 
      (*       check_polarity uenv; *)
      if not ((Fable.is_empty mtv.ftype) && Fable.is_empty t.ftype) then
        raise (Unify_failure "Unable to unify fable qualifiers")
      else
        match mtv.ltype with 
            `MetaTypeVar point as mtv_l -> 
              begin
                (match (Unionfind.find point) with
                   | Rigid l -> 
                       raise (Unify_failure ("Couldn't unify the rigid type variable "^ string_of_int l ^" with the type "^ string_of_datatype t))
                   | Flexible var ->
                       begin
                         if var_is_free_in_type var t then
                           begin
                             Debug.if_set (show_recursion)
                               (fun () -> "rec intro3 ("^string_of_int var^","^string_of_datatype t^")");
                             if not (Fable.is_empty t.ftype) then 
                               raise (Unify_failure "Unification of fable qualifiers in recursive types is not currently supported");
                             let point' = Unionfind.fresh (Body t)
                             in
                               rec_intro uenv point' (var, t);
                               Unionfind.union point point'
                           end
                         else
                           (Debug.if_set (show_recursion) (fun () -> "non-rec intro (" ^ string_of_int var ^ ")");
                            Unionfind.change point (Body t))
                       end;
                       Subst.empty
                   | Recursive (var, t') ->
                       Debug.if_set (show_recursion) (fun () -> "rec single (" ^ (string_of_int var) ^ ")");
                       if not (Fable.is_empty t.ftype) then 
                         raise (Unify_failure "Unification of fable qualifiers in recursive types is not currently supported")
                       else
                         begin
                           if is_unguarded_recursive mtv_l then 
                             raise (Unify_failure ("Couldn't unify the unguarded recursive type "^
                                                     string_of_datatype mtv ^
                                                     " with the non-recursive type "^ string_of_datatype t))
                           else                   
                             unify_rec uenv ((var, t'), t)
                         end
                           (* It's tempting to try to do this, but it isn't sound
                              as point may appear inside t
                              
                              Unionfind.change point t;
                           *)
                   | Body t' -> 
                       if not (Fable.is_empty t.ftype) then 
                         raise (Unify_failure "Failed to unify fable qualifiers")
                       else
                         unify' uenv (t', t))
              end
          | _ -> failwith "Impossible" in
      
    let unify_app uenv app t : Subst.subst = match app.ltype with 
      | `Application (s, ts) -> 
          let vars, alias = lookup_alias (s, ts) uenv in
            begin
              match alias.ltype with
                | `Primitive `Abstract ->
                    raise (Unify_failure
                             ("Cannot unify abstract type '"^ string_of_datatype app  ^"' with type '"^string_of_datatype t^"'"))
                | _ ->
                    unify' uenv (instantiate_alias (vars, alias) ts, t)
            end
      | _ -> failwith "Impossible" in

    let check_empty_quals (t1,t2) =
      if not (Fable.is_empty t1.ftype && Fable.is_empty t2.ftype) then
	    raise (Unify_failure "Could not unify fable qualifiers") in
	  (*         match (concrete_type t1).ltype, (concrete_type t2).ltype with (\* fable quals were not completely unified; *\) *)
	  (*             `MetaTypeVar _, _ (\* only possibility is for the remaining quals to be unified with an mtv *\) *)
	  (*           | _, `MetaTypeVar _ -> () *)
	  (*           | _ -> raise (Unify_failure "Could not unify fable qualifiers") *)
	  (*       else () in *)

    let assert_empty_quals (t1,t2) = 
      if not (Fable.is_empty t1.ftype && Fable.is_empty t2.ftype) then
        raise (Unify_failure "Could not unify fable qualifiers") 
      else () in
      
    let unify_l uenv (t1, t2) : Subst.subst =
      check_polarity uenv;
      (Debug.if_set (show_unification) (fun () -> "Unifying "^string_of_datatype t1^" with "^string_of_datatype t2);
       check_empty_quals (t1,t2);
       let sigma = 
	     (match (t1.ltype, t2.ltype) with
            | `Not_typed, _ | _, `Not_typed -> failwith "Internal error: `Not_typed' passed to `unify'"
            | `Primitive x, `Primitive y when x = y -> Subst.empty
            | `MetaTypeVar lpoint, `MetaTypeVar rpoint -> 
		        if Unionfind.equivalent lpoint rpoint then
		          ()
		        else
                  begin
                    let unify_flex_t ((t1 : datatype), point1, var) ((t2:datatype), point2) =
		              (if var_is_free_in_type var t2 then (* no fable types remain to be unified *)
                         (Debug.if_set (show_recursion) (fun () -> "rec intro1 (" ^ (string_of_int var) ^ ")");
                          rec_intro uenv rpoint (var, Types.concrete_type t2))
		               else
                         Unionfind.union point1 point2)                    
                    in
                      match (Unionfind.find lpoint, Unionfind.find rpoint) with
			            | Rigid l, Rigid r ->
                            if l <> r then 
                              raise (Unify_failure ("Rigid type variables "^ string_of_int l ^" and "^ string_of_int r ^" do not match"))
                            else 
                              failwith "Unreachable case reached"
				                (*                          Unionfind.union lpoint rpoint (\* Question : Is this case ever reachable? *\) *)
			            | Flexible _, Flexible _ ->
                            Unionfind.union lpoint rpoint

			            | Flexible var, _ -> unify_flex_t (t1, lpoint, var) (t2, rpoint)
			            | _, Flexible var -> unify_flex_t (t2, rpoint, var) (t1, lpoint)
                            
			            | Rigid l, _ ->
                            raise (Unify_failure ("Couldn't unify the rigid type variable "^
                                                    string_of_int l ^" with the type "^ string_of_datatype t2))
			            | _, Rigid r ->
                            raise (Unify_failure ("Couldn't unify the rigid type variable "^
                                                    string_of_int r ^" with the type "^ string_of_datatype t1))
			            | Recursive (lvar, t), Recursive (rvar, t') ->
                            assert(lvar <> rvar);
                            assert_empty_quals (t1,t2);
                            Debug.if_set (show_recursion)
                              (fun () -> "rec pair (" ^ (string_of_int lvar) ^ "," ^ (string_of_int rvar) ^")");
                            begin
                              if is_unguarded_recursive (`MetaTypeVar lpoint) then
				                begin
                                  if not (is_unguarded_recursive (`MetaTypeVar rpoint)) then
                                    raise (Unify_failure ("Couldn't unify the unguarded recursive type "^
                                                            string_of_datatype t1 ^
                                                            " with the guarded recursive type "^ string_of_datatype t2))
				                end
                              else if is_unguarded_recursive (`MetaTypeVar lpoint) then
				                raise (Unify_failure ("Couldn't unify the unguarded recursive type "^
							                            string_of_datatype t2 ^
							                            " with the guarded recursive type "^ string_of_datatype t1));
                              unify_rec2 uenv ((lvar, t), (rvar, t'))
                            end;
                            Unionfind.union lpoint rpoint
                              
			            | Recursive (var, t'), Body t ->
                            assert_empty_quals (t1,t2);
                            Debug.if_set (show_recursion) (fun () -> "rec left (" ^ (string_of_int var) ^ ")");
                            begin
                              if is_unguarded_recursive (`MetaTypeVar lpoint) then
				                raise (Unify_failure ("Couldn't unify the unguarded recursive type "^
							                            string_of_datatype t1  ^
							                            " with the non-recursive type "^ string_of_datatype t2))
                              else                   
				                unify_rec uenv ((var, t'), t)
                            end;
                            Unionfind.union rpoint lpoint
                              
			            | Body t, Recursive (var, t') ->
                            assert_empty_quals (t1,t2);
                            Debug.if_set (show_recursion) (fun () -> "rec right (" ^ (string_of_int var) ^ ")");
                            begin
                              if is_unguarded_recursive (`MetaTypeVar rpoint) then
				                raise (Unify_failure ("Couldn't unify the unguarded recursive type "^
							                            string_of_datatype t2 ^
							                            " with the non-recursive type "^ string_of_datatype t1))
                              else                   
				                unify_rec (flip_polarity uenv) ((var, t'), t)
                            end;
                            Unionfind.union lpoint rpoint
			            | Body t, Body t' -> 
                            assert_empty_quals (t1,t2);
                            unify' uenv (t, t'); 
                            Unionfind.union lpoint rpoint
                  end;
		        Subst.empty
            | `MetaTypeVar point, t -> unify_mtv uenv t1 t2
            | t, `MetaTypeVar point -> unify_mtv (flip_polarity uenv) t2 t1 
            | `Function (lfrom, lm, lto), `Function (rfrom, rm, rto) ->
		        let phl, noptl = Fable.get_name_bindings lfrom.ftype in
		        let phr, noptr = Fable.get_name_bindings rfrom.ftype in
		        let lfrom, rfrom = Fable.strip_name_bindings_type lfrom, Fable.strip_name_bindings_type rfrom in
		        let sigma0 = 
                  if (List.length phl <> List.length phr) then
                    raise (Unify_failure "Couldn't unify function types with different number of phantom variables")
                  else
                    List.fold_left2 (fun sub nl nr -> 
                                       Subst.extend (nr, Variable(nl, `T(dummy_position, Fable.fresh_label_type(), None))) sub)
                      Subst.empty phl phr in
		        let sigma0 = 
                  match noptl, noptr with
                      None, None -> sigma0
                    | Some nl, Some nr -> Subst.extend (nr, Variable(nl, `T(dummy_position, Fable.fresh_label_type(), None))) sigma0 
                    | _ -> raise (Unify_failure "Couldn't unify function types with incompatible name bindings") in
		        let uenv = clear_expression uenv in
		        let sigma1 = unify' uenv (lfrom, Subst.apply sigma0 rfrom) in
		        let sigma01 = Subst.union sigma0 sigma1 in
		        let sigma2 = unify' uenv (lto, Subst.apply sigma01 rto) in
		        let sigma012 = Subst.union sigma01 sigma2 in
		        let sigma3 = unify' uenv (lm, Subst.apply sigma012 rm) in
                  Subst.union sigma012 sigma3 
            | `Record l, `Record r -> unify_record_rows uenv (l, r)
            | `Variant l, `Variant r -> unify_variant_rows uenv (l, r)
            | `Table (lr, lw), `Table (rr, rw) ->
		        let uenv = clear_expression uenv in
		        let sigma = unify' uenv (lr, rr) in
		        let sigma' = unify' uenv (lr, Subst.apply sigma rr) in
		          Subst.union sigma sigma'
            | `Application (ls, lts), `Application (rs, rts) when ls = rs -> 
		        let uenv = clear_expression uenv in 
		          List.fold_left2 (fun sigma lt rt -> let sigma' = unify' uenv (lt, Subst.apply sigma rt) in Subst.union sigma sigma') Subst.empty lts rts
            | `Application (ls, lts), `Application (rs, rts) ->
		        let lvars, lalias = lookup_alias (ls, lts) uenv
		        and rvars, ralias = lookup_alias (rs, rts) uenv in
                  begin
                    match lalias.ltype, ralias.ltype with
                      | `Primitive `Abstract, `Primitive `Abstract ->
                          raise (Unify_failure
                                   ("Cannot unify abstract type '"^string_of_datatype t1^
                                      "' with abstract type '"^string_of_datatype t2^"'"))
                      | `Primitive `Abstract, _ ->
                          unify' uenv (t1, instantiate_alias (rvars, ralias) rts)
                      | _, `Primitive `Abstract ->
                          unify' uenv (instantiate_alias (lvars, lalias) lts, t2)
                      | _, _ ->
                          unify' uenv (instantiate_alias (lvars, lalias) lts,
                                       instantiate_alias (rvars, ralias) rts)
                  end
            | `Application (s, ts), t  -> unify_app uenv t1 t2
            | t, `Application (s, ts) -> unify_app (flip_polarity uenv) t2 t1
            | _, _ ->
		        raise (Unify_failure ("Couldn't match "^ string_of_datatype t1 ^" against "^ string_of_datatype t2))) in
	     Debug.if_set (show_unification) (fun () -> "Unified types: " ^ string_of_datatype t1);
	     sigma) in
    let unify_ftype uenv (tf1, tf2) : ((fable_type * fable_type) * Subst.subst) = 
      check_polarity uenv;
      let fail msg = raise (Unify_failure ("Could not unify types; incompatible fable qualifiers: " ^ 
					                         (Expression.DerivingSyntax.Show_fable_type.show tf1) ^ " <> " ^
					                         (Expression.DerivingSyntax.Show_fable_type.show tf2) ^ " : " )) in
      let rec aux sigma = function 
          `Name _, `Name _
        | `Phantom _, `Phantom _ -> failwith "Unexpected top-level name binding in fable type qualifier"
	        
	    |  q1, q2 when q1 = q2 -> (`Empty, `Empty), sigma 
	         
	    | `Kind k1, `Kind k2 when Fable.sub_kind (k1,k2) -> (`Empty, `Empty), sigma 

	    | (`Kind _ as tf1), `Label(`Inferred pt, tf2) -> 
	        Unionfind.change pt `Nothing;
	        aux sigma (tf1, tf2)

	    | (`KindVar _ as tf1), `Label(`Inferred pt, tf2) -> 
	        Unionfind.change pt `Nothing;
	        aux sigma (tf1, tf2)

	    | `KindVar kv1, `KindVar kv2 when Unionfind.equivalent kv1 kv2 -> 
	        (`Empty, `Empty), sigma

	    | `KindVar kv1, `KindVar kv2 ->
	        begin
	          match Unionfind.find kv1, Unionfind.find kv2 with
		          `FixedKind k1, `FlexKind (_, k2) -> 
		            if  Fable.sub_kind (Fable.kind_qual k1,k2) then () 
		            else fail()
			          
		        | `FlexKind (_, k1), `FixedKind k2 -> 
		            if  Fable.sub_kind (Fable.kind_qual k2,k1) then
			          Unionfind.change kv1 (`FixedKind k2)
		            else fail()
			          
		        | `FlexKind (_, k1), `FlexKind (_, k2) -> 
		            if (Fable.sub_kind (k1,k2)) then
		              Unionfind.union kv2 kv1 
		            else 
			          Unionfind.union kv1 kv2 
			            
		        | `FixedKind k1, `FixedKind k2 -> 
		            let _, sigma' = aux sigma (k1, k2) in
			          if (sigma <> sigma') then fail (k1,k2) 
	        end;
	        (`Empty, `Empty), sigma
	          
	    | fq1, `KindVar kv2 -> 
	        (match Unionfind.find kv2 with 
	           | `FixedKind fq2 -> aux sigma (fq1, fq2)
	           | `FlexKind (_, kb) when Fable.sub_kind (Fable.kind_qual fq1, kb) ->
		           Unionfind.change kv2 (`FixedKind fq1); (`Empty, `Empty), sigma
	           | _ -> fail ())
	          
	    | `KindVar kv1, fq2 -> 
	        let _ = match Unionfind.find kv1 with 
	          | `FixedKind fq1 -> aux sigma (fq1, fq2)
	          | `FlexKind (_, kb) when Fable.sub_kind (kind_qual fq2, kb) ->
		          Unionfind.change kv1 (`FixedKind fq2); (`Empty, `Empty), sigma
	          | _ -> fail () in
	          (`Empty, `Empty), sigma
	            
        | `Label (e1opt, tf1), `Label(e2opt, tf2) ->
            let sigma' = match e1opt, e2opt with 
	          | `Nothing, `Nothing -> sigma
	          | `Nothing, `Required e2 -> 
		          (match uenv.exp with 
		               None -> fail ()
		             | Some e1 -> (* subtype e:lab <: e:lab~e *)
			             let sigma' = Expequiv.unify uenv.phantoms uenv.omega (e1, e2) in
			               Subst.union sigma sigma' )

	          | `Required _, `Nothing -> sigma

	          | `Nothing, `Inferred pt ->
		          Unionfind.change pt `Nothing; sigma
	          | `Inferred pt, `Nothing -> 
		          if not (uenv.subtype) then 
		            Unionfind.change pt `Nothing; 
		          sigma
		            
              | `Inferred p1, `Inferred p2 -> 
		          (match Unionfind.find p1, Unionfind.find p2 with
		               `Nothing, `Nothing -> sigma
		             | `Something e, `Nothing -> 
			             if not uenv.subtype then
			               Unionfind.change p1 `Nothing; 
			             sigma
		             | `Nothing, `Something e -> Unionfind.change p2 `Nothing; sigma
		             | `Something e1, `Something e2 -> 
			             try
			               let sigma' = Expequiv.unify uenv.phantoms uenv.omega (e1,e2) in
			                 Subst.union sigma sigma'
			             with
			                 Expequiv.Equiv_failure _ -> 
			                   Unionfind.change p2 `Nothing; 
			                   if not uenv.subtype then
				                 Unionfind.change p1 `Nothing;
			                   sigma)
		            
              | `Inferred pt, `Required(e2) -> 
		          (match Unionfind.find pt with
		               `Nothing -> fail ()
		             | `Something e1 -> 
			             let sigma' = Expequiv.unify uenv.phantoms uenv.omega (e1, e2) in
			             Subst.union sigma sigma')
                    
	          | `Required(e1), `Required(e2) -> 
		          let sigma' = Expequiv.unify uenv.phantoms uenv.omega (e1, e2) in
		          Subst.union sigma sigma'
		            
	          | `Required(e1), `Inferred(e2) -> raise Impos
                  
	        in
            let tf1' = Subst.apply_fq sigma' tf1 in
            let tf2' = Subst.apply_fq sigma' tf2 in
              aux sigma' (tf1', tf2')
		        
	    | `Label(_, tf1), (`Kind _ as tf2) -> aux sigma (tf1, tf2)
            (* 	| `Label(_, tf1), (`KindVar _ as tf2) -> aux sigma (tf1, tf2) *)

	    | (`Kind _ as tf1), `Label(`Nothing, tf2) -> aux sigma (tf1, tf2)
            (* 	| (`KindVar _ as tf1), `Label(`Nothing, tf2) -> aux sigma (tf1, tf2) *)
	        
        | `Labeling(e1, tf1), `Labeling(e2, tf2) -> 
            let sigma' = Expequiv.unify uenv.phantoms uenv.omega (e1, e2) in
            let sigma' = Subst.union sigma sigma' in
            let tf1' = Subst.apply_fq sigma' tf1 in
            let tf2' = Subst.apply_fq sigma' tf2 in
              aux sigma' (tf1', tf2')

        | `Empty,  tf2 ->
	        (match tf2 with
		         `Kind MKind 
	           | `Kind UKind -> (`Empty, `Empty), sigma 
 	           | _ -> fail ())

        |  tf1, `Empty -> 
	         (match tf1 with
		          `Kind _ -> (`Empty, `Empty), sigma 
		        | `Label(_, tf1') -> aux sigma (tf1', `Empty)
 		        | _ -> fail())
	           
  	    | _ -> fail () in
	    try 
          aux Subst.empty (tf1, tf2)
	    with 
	        Expequiv.Equiv_failure msg -> raise (Unify_failure msg) in
      fun uenv (t1,t2) -> 
	    match uenv.polarity with 
	        `Neg -> unify' (flip_polarity uenv) (t2, t1)
	      | `Pos -> 
              let (f1', f2'), sigma = unify_ftype uenv (t1.ftype,t2.ftype) in
              let t1' = Subst.apply sigma {ltype = t1.ltype; ftype = f1'} in
              let t2' = {ltype = t2.ltype; ftype = f2'} in 
		        assert_msg (list_intersect (Subst.domain sigma) (Fable.free_names uenv.omega t2') = []) "Domain sigma contains unexpected names"; 
		        let sigma' = unify_l uenv (t1',t2') in
		          Subst.union sigma sigma' 
		            
  and unify_variant_rows : unify_env -> (row * row) -> Subst.subst = 
    fun unify_env (lrow, rrow) -> 
      
      (* 
         [NOTE]

         - All calls to fail_on_absent_fields are currently disabled,
         as under the current model absent fields have
         to be allowed in closed rows (although they're ignored).

         - There's no way of getting rid of absent variables as they're stored in the field
         environment rather than the row variable (good argument for moving them into the
         row variable).
      *)
      (*
        let fail_on_absent_fields field_env =
        StringMap.iter
        (fun _ -> function
        | `Present _ -> ()
        | `Absent ->
        failwith "Internal error: closed row with absent variable"
        ) field_env in
      *)

      let is_unguarded_recursive row =
        let rec is_unguarded rec_rows (field_env, row_var) =
          StringMap.is_empty field_env &&
            (match Unionfind.find row_var with
               | ClosedRow
               | FlexibleRow _
               | RigidRow _ -> false
               | RecursiveRow (var, row) when IntSet.mem var rec_rows -> true
               | RecursiveRow (var, row) -> is_unguarded (IntSet.add var rec_rows) row
               | BodyRow row -> is_unguarded rec_rows row)
        in
          is_unguarded IntSet.empty row in
        
      (* extend_field_env traversal_env extending_env
         extends traversal_env with all the fields in extending_env

         Matching `Present fields are unified.

         Any fields in extending_env, but not in traversal_env are
         added to an extension environment which is returned.
      *)
      let extend_field_env
          (uenv : unify_env)
          (traversal_env : field_spec_map)
          (extending_env : field_spec_map) =
        StringMap.fold
          (fun label field_spec (extension, sigma) ->
             if StringMap.mem label extending_env then
               (match field_spec, (StringMap.find label extending_env) with
                  | `Present t, `Present t' ->
                      let sigma' = unify' uenv (t, Subst.apply sigma t') in
                        extension, Subst.union sigma (Subst.restrict sigma' uenv.phantoms)
                  | `Absent, `Absent ->
                      extension, sigma
                  | `Present _, `Absent
                  | `Absent, `Present _ ->
                      raise (Unify_failure ("Rows\n "^ string_of_row lrow
                                            ^"\nand\n "^ string_of_row rrow
                                            ^"\n could not be unified because they have conflicting fields"))
               )
             else
               StringMap.add label field_spec extension, sigma
          ) traversal_env (StringMap.empty, Subst.empty) in
	    
      let unify_compatible_field_environments uenv (field_env1, field_env2) =
        extend_field_env uenv field_env1 field_env2 in
	    
      (* introduce a recursive row
         give an error if it is non-well-founded and
         non-well-founded type inference is switched off
      *)
      let rec_row_intro uenv point (var, row) =
	    let alias_env = Typeenv.alias_env uenv.omega in
          if occurs_check_row alias_env var row then
            Unionfind.change point (RecursiveRow (var, row))
          else
            raise (Unify_failure ("Cannot unify row variable "^string_of_int var^" with row "^string_of_row row^
                                    " because "^
                                    match Settings.get_value infer_recursive_types with
					                  | "guarded" -> "the row variable occurs unguarded inside the row"
					                  | "positive" -> "the row variable occurs in a negative position inside the row"
					                  | _ -> assert false)) in
	    
      (*
        unify_row_var_with_row rec_env (row_var, row)
        attempts to unify row_var with row
        
        However, row_var may already have been instantiated, in which case
        it is unified with row.
      *)
      let unify_row_var_with_row : unify_env -> row_var * row -> unit =
        fun uenv (row_var, ((extension_field_env, extension_row_var) as extension_row)) ->
          (* unify row_var with `RowVar None *)
          let close_empty_row_var : row_var -> unit = fun point ->
            match Unionfind.find point with
              | ClosedRow -> ()
              | FlexibleRow _ ->
                  Unionfind.change point ClosedRow
              | RigidRow _ ->
                  raise (Unify_failure ("Closed row var cannot be unified with rigid row var\n"))
              | _ -> assert false in

          (* unify row_var with `RigidRowVar var *)
          let rigidify_empty_row_var var : row_var -> unit = fun point ->
            match Unionfind.find point with
              | ClosedRow ->
                  raise (Unify_failure ("Rigid row var cannot be unified with empty closed row\n"))
              | FlexibleRow _ ->
                  Unionfind.change point (RigidRow var)
              | RigidRow var' when var=var' -> ()
              | RigidRow var' ->
                  raise (Unify_failure ("Incompatible rigid row variables cannot be unified\n"))
              | _ -> assert false in

          let rec extend = function
            | point ->
                (* point should be a row variable *)
                let row_var = Unionfind.find point in
                  match row_var with
                    | ClosedRow ->
                        if is_empty_row extension_row then
                          close_empty_row_var extension_row_var
                        else
                          raise (Unify_failure ("Closed row cannot be extended with non-empty row\n"
                                                ^string_of_row extension_row))
                    | RigidRow var ->
                        if is_empty_row extension_row then
                          rigidify_empty_row_var var extension_row_var
                        else
                          raise (Unify_failure ("Rigid row variable cannot be unified with non-empty row\n"
                                                ^string_of_row extension_row))
                    | FlexibleRow var ->
                        if TypeVarSet.mem var (free_row_type_vars extension_row) then
                          rec_row_intro uenv point (var, extension_row)
                        else if StringMap.is_empty extension_field_env then
                          match extension_row_var with
                            | point' ->
                                Unionfind.union point point'
                        else
                          Unionfind.change point (BodyRow extension_row)
                    | RecursiveRow _ ->
                        ignore (unify_variant_rows uenv ((StringMap.empty, point), extension_row)) (* TODO: check this *)
                    | BodyRow row ->
                        ignore (unify_variant_rows uenv (row, extension_row))  (* TODO: check this *)
          in
            extend row_var in
	    

      (* 
         matching_labels (big_field_env, small_field_env)
         return the set of labels that appear in both big_field_env and small_field_env

         precondition: big_field_env contains small_field_env
      *)
      let matching_labels : field_spec_map * field_spec_map -> StringSet.t = 
        fun (big_field_env, small_field_env) ->
          StringMap.fold (fun label _ labels ->
                            if StringMap.mem label small_field_env then
                              StringSet.add label labels
                            else
                              labels) big_field_env StringSet.empty in

      let row_without_labels : StringSet.t -> row -> row =
        fun labels (field_env, row_var) ->
          let restricted_field_env =
            StringSet.fold (fun label field_env ->
                              StringMap.remove label field_env) labels field_env
          in
            (restricted_field_env, row_var) in

      (*
        register a recursive row in the rec_env environment
        
        return:
        None if the recursive row already appears in the environment
        Some rec_env, otherwise, where rec_env is the updated environment
      *)
      let register_rec_row (wrapped_field_env, unwrapped_field_env, rec_row, unwrapped_row') : unify_env -> unify_env option =
        fun uenv -> (* ((rec_types, rec_rows, alias_env) as rec_env) -> *)
          match rec_row with
            | Some row_var ->
                begin
                  match Unionfind.find row_var with
                    | RecursiveRow (var, _) ->
                        let restricted_row = row_without_labels (matching_labels (unwrapped_field_env, wrapped_field_env)) unwrapped_row' in
                        let rs =
                          if IntMap.mem var uenv.rec_rows then
                            IntMap.find var uenv.rec_rows
                          else
                            [(StringMap.empty, row_var)]
                        in
                          if List.exists (fun r ->
                                            eq_rows (r, restricted_row)) rs then
                            None
                          else
                            Some (add_rec_rows var (restricted_row::rs) uenv)
                    | _ -> assert false
                end
            | None ->
                Some uenv in

      (*
        register two recursive rows and return None if one of them is already in the environment
      *)
      let register_rec_rows p1 p2 : unify_env -> unify_env option = fun uenv ->
        let uenv' = register_rec_row p1 uenv in
          match uenv' with
            | None -> None
            | Some uenv -> register_rec_row p2 uenv in

      let unify_both_rigid_with_rec_env uenv ((lfield_env, _ as lrow), (rfield_env, _ as rrow)) =
        let get_present_labels (field_env, row_var) =
          let rec get_present' rec_vars (field_env, row_var) =
            let top_level_labels = 
              StringMap.fold (fun label field_spec labels ->
                                match field_spec with
                                  | `Present _ -> StringSet.add label labels
                                  | `Absent -> labels) field_env StringSet.empty
            in
              StringSet.union top_level_labels 
                (match Unionfind.find row_var with
                   | RecursiveRow (var, body) when (not (IntSet.mem var rec_vars)) ->
                       get_present' (IntSet.add var rec_vars) body
                   | _ -> StringSet.empty) in
            get_present' IntSet.empty (field_env, row_var) in
          
        let fields_are_compatible (lrow, rrow) =
          (StringSet.equal (get_present_labels lrow) (get_present_labels rrow)) in
          
        let (lfield_env', lrow_var') as lrow', lrec_row = unwrap_row lrow in
        let (rfield_env', rrow_var') as rrow', rrec_row = unwrap_row rrow in
          (*
            fail_on_absent_fields lfield_env;
            fail_on_absent_fields rfield_env;
          *)
          match Unionfind.find lrow_var', Unionfind.find rrow_var' with
            | RigidRow lvar, RigidRow rvar when lvar <> rvar ->
                raise (Unify_failure ("Rigid rows\n "^ string_of_row lrow
                                      ^"\nand\n "^ string_of_row lrow
                                      ^"\n could not be unified because they have distinct rigid row variables"))
            | _, _ ->
                if fields_are_compatible (lrow', rrow') then
                  let uenv' =
                    (register_rec_rows
                       (lfield_env, lfield_env', lrec_row, rrow')
                       (rfield_env, rfield_env', rrec_row, lrow')
                       uenv)
                  in
                    match uenv' with
                      | None -> Subst.empty
                      | Some uenv ->
                          snd (unify_compatible_field_environments uenv (lfield_env', rfield_env'))
                else
                  raise (Unify_failure ("Rigid rows\n "^ string_of_row lrow
                                        ^"\nand\n "^ string_of_row rrow
                                        ^"\n could not be unified because they have different fields")) in

      let unify_both_rigid uenv = unify_both_rigid_with_rec_env uenv in

      let unify_one_rigid uenv ((rigid_field_env, _ as rigid_row), (open_field_env, _ as open_row)) =
        let (rigid_field_env', rigid_row_var') as rigid_row', rigid_rec_row = unwrap_row rigid_row in
        let (open_field_env', open_row_var') as open_row', open_rec_row = unwrap_row open_row in 
          (* check that the open row contains no extra fields *)
          StringMap.iter
            (fun label field_spec ->
               if (StringMap.mem label rigid_field_env') then
                 ()
               else
                 match field_spec with
                   | `Present _ ->
                       raise (Unify_failure
                                ("Rows\n "^ string_of_row rigid_row
                                 ^"\nand\n "^ string_of_row open_row
                                 ^"\n could not be unified because the former is rigid"
                                 ^" and the latter contains fields not present in the former"))
                   | `Absent -> ()
            ) open_field_env';
          
          (* check that the closed row contains no absent fields *)
          (*          fail_on_absent_fields closed_field_env; *)
          
          let uenv' =
            (register_rec_rows
               (rigid_field_env, rigid_field_env', rigid_rec_row, open_row')
               (open_field_env, open_field_env', open_rec_row, rigid_row')
               uenv)
          in
            match uenv' with
              | None -> Subst.empty
              | Some uenv ->
                  let open_extension, sigma = extend_field_env uenv rigid_field_env' open_field_env' in
                  let _ = unify_row_var_with_row uenv (open_row_var', (open_extension, rigid_row_var')) in
		            sigma in

      let unify_both_open uenv ((lfield_env, _ as lrow), (rfield_env, _ as rrow)) =
        let (lfield_env', lrow_var') as lrow', lrec_row = unwrap_row lrow in
        let (rfield_env', rrow_var') as rrow', rrec_row = unwrap_row rrow in
        let uenv' =
          (register_rec_rows
             (lfield_env, lfield_env', lrec_row, rrow')
             (rfield_env, rfield_env', rrec_row, lrow')
             uenv)
        in
          match uenv' with
            | None -> Subst.empty
            | Some uenv ->
                if (get_row_var lrow = get_row_var rrow) then     
                  unify_both_rigid_with_rec_env uenv ((lfield_env', Unionfind.fresh ClosedRow),
                                                      (rfield_env', Unionfind.fresh ClosedRow))
                else
                  begin
                    let fresh_row_var = fresh_row_variable() in          
                      (* each row can contain fields missing from the other; 
                         thus we call extend_field_env once in each direction *)
                    let rextension, sigma =
                      extend_field_env uenv lfield_env' rfield_env' in
                      (* [NOTE]
                         extend_field_env may change rrow_var' or lrow_var', as either
                         could occur inside the body of lfield_env' or rfield_env'
                      *)
                      unify_row_var_with_row uenv (rrow_var', (rextension, fresh_row_var));
                      let lextension, sigma' = extend_field_env uenv rfield_env' lfield_env' in
                      let _ = unify_row_var_with_row uenv (lrow_var', (lextension, fresh_row_var)) in
			            Subst.union sigma sigma'
                  end in
        
      (* report an error if an attempt is made to unify
         an unguarded recursive row with a row that is not
         unguarded recursive
      *)
      let check_unguarded_recursion lrow rrow =      
        if is_unguarded_recursive lrow then
          (if not (is_unguarded_recursive rrow) then
             raise (Unify_failure
                      ("Could not unify unguarded recursive row"^ string_of_row lrow
                       ^"\nwith row "^ string_of_row rrow)))
        else if is_unguarded_recursive rrow then
          raise (Unify_failure
                   ("Could not unify unguarded recursive row"^ string_of_row rrow
                    ^"\nwith row "^ string_of_row lrow)) in

	  let uenv = clear_expression unify_env in
        Debug.if_set (show_row_unification) (fun () -> "Unifying row: " ^ (string_of_row lrow) ^ " with row: " ^ (string_of_row rrow));
	    
        let sigma = 
	      check_unguarded_recursion lrow rrow;
	      
	      if is_rigid_row lrow then
	        if is_rigid_row rrow then
		      unify_both_rigid uenv (lrow, rrow)
	        else
		      unify_one_rigid uenv (lrow, rrow)
	      else if is_rigid_row rrow then
	        unify_one_rigid (flip_polarity uenv) (rrow, lrow)           
	      else
	        unify_both_open uenv (lrow, rrow)
        in
	      Debug.if_set (show_row_unification)
	        (fun () -> "Unified rows: " ^ (string_of_row lrow) ^ " and: " ^ (string_of_row rrow));
	      sigma
            
  and unify_record_rows uenv (lrow, rrow) = 
    let unify_dependent_rows uenv (lrow, rrow) = 
	  
      let rec get_field_expression label = function
	    | Some (Record_intro (fmap, ext, _)) -> 
	        if StringMap.mem label fmap then
	          Some (StringMap.find label fmap)
	        else
	          get_field_expression label ext
	    | _ -> None in
	    
      (* Precondition : must be an unwrapped row *)
      let get_present_labels (field_env, row_var) =
        let rec get_present' rec_vars (field_env, row_var) =
          let top_level_labels = 
            StringMap.fold (fun label field_spec labels ->
                              match field_spec with
                                | `Present _ -> StringSet.add label labels
                                | `Absent -> labels) field_env StringSet.empty
          in
            StringSet.union top_level_labels 
              (match Unionfind.find row_var with
                 | RecursiveRow (var, body) when (not (IntSet.mem var rec_vars)) ->
                     raise (Unify_failure "Unexpected recursive row variable in a record with fable qualifiers")
                 | _ -> StringSet.empty) in
          get_present' IntSet.empty (field_env, row_var) in

      let unify_field_envs
          (uenv : unify_env)
          (labels : string list)
          (fsml, fsmr) = 
        (*         let labels = List.sort compare labels in *)
        let get_fsm_types label = 
          match StringMap.find label fsml, StringMap.find label fsmr with
              `Present t, `Present s -> t,s
            | _ -> raise (Unify_failure "Unexpected `Absent field qualifier") in
        let unify_field (sigma_in, sig_inv) label =
          let (ti, si) = get_fsm_types label in
          let pi, xi = Fable.get_name_bindings ti.ftype in
          let qi, yi = Fable.get_name_bindings si.ftype in
          let _ = match pi, qi with 
              [], [] -> ()
            | _ -> raise (Unify_failure "Unexpected phantom variables in record field type") in
          let ti' = Fable.strip_name_bindings_type ti in
          let si' = Fable.strip_name_bindings_type si in
	      let field_exp = get_field_expression label uenv.exp in
          let sigma' = unify' ({uenv with exp=field_exp}) (ti', Subst.apply sigma_in si') in
	      let sigma' = Subst.union sigma_in sigma' in
          let sigma_out, sig_inv' = match xi, yi with
            | Some(xi), Some(yi) -> 
                Subst.extend (yi, Variable(xi, `T(dummy_position, ti', None))) sigma', 
		        Subst.extend (xi, Variable(yi, `T(dummy_position, ti', None))) sig_inv
	        | None, Some(yi) -> (* Intro dependent product *)
		        let lte = try Fable.get_label_type_expr ti'.ftype with Not_found -> None in
		          (match lte, field_exp with 
		               None, None -> 
			             raise (Unify_failure "Unable to infer dependent product introduction; try using \"var v = pack e as t\"")
 		             | Some e, _ 
		             | _, Some e -> Subst.extend (yi, e) sigma', sig_inv)
	        | _ -> sigma', sig_inv in
            sigma_out, sig_inv' in
          List.fold_left unify_field (Subst.empty, Subst.empty) labels in
	    
      (************* Record unification rule for both rigid rows:

                     sigma_0 = {}            
                     p, Om |- ti < sigma_{i-1} (si) : sigma_i'
                     sigma_i = sigma_0, ..., sigma_{i-1}, sigma_i', (yi -> xi)
                     ------------------------------------------------------------------------
                     p, Om |- (f1: ~x1:t1, ..., fn :tn) < (f1: ~y1:s1, ..., fn :sn) 
      *************)                                    

      let unify_both_rigid uenv ((lfield_env, _ as lrow), (rfield_env, _ as rrow)) =          
        let fields_are_compatible (lrow, rrow) =
          let llabs = get_present_labels lrow in
          let rlabs = get_present_labels rrow in
            if StringSet.equal llabs rlabs then
              List.map fst (Fable.row_labels_inorder ~filter_absent:true (fst rrow))
            else 
              raise (Unify_failure ("Rigid rows\n "^ string_of_row lrow
                                    ^"\nand\n "^ string_of_row rrow
                                    ^"\n could not be unified because they have different fields")) in
          
        let (lfield_env', lrow_var') as lrow', lrec_row = unwrap_row lrow in
        let (rfield_env', rrow_var') as rrow', rrec_row = unwrap_row rrow in
        let _ = match lrec_row, rrec_row with
            None, None -> ()
          | _ -> raise (Unify_failure "Unexpected recursive row variable in a record with fable qualifiers") in
	      
          match Unionfind.find lrow_var', Unionfind.find rrow_var' with
            | RigidRow lvar, RigidRow rvar when lvar <> rvar ->
                raise (Unify_failure ("Rigid rows\n "^ string_of_row lrow
                                      ^"\nand\n "^ string_of_row lrow
                                      ^"\n could not be unified because they have distinct rigid row variables"))
            | _, _ ->
                let fields = fields_are_compatible (lrow', rrow') in
                let sigma, _ = unify_field_envs uenv fields (lfield_env', rfield_env') in
		          sigma
      in
	    
      let unify_row_var_with_row : unify_env -> row_var * row -> unit = 
	    fun uenv (row_var, ((extension_field_env, extension_row_var) as extension_row)) ->
	      (* unify row_var with `RowVar None *)
	      let close_empty_row_var : row_var -> unit = fun point ->
	        match Unionfind.find point with
              | ClosedRow -> ()
              | FlexibleRow _ ->
		          Unionfind.change point ClosedRow
              | RigidRow _ ->
		          raise (Unify_failure ("Closed row var cannot be unified with rigid row var\n"))
              | _ -> assert false in

	      (* unify row_var with `RigidRowVar var *)
	      let rigidify_empty_row_var var : row_var -> unit = fun point ->
	        match Unionfind.find point with
              | ClosedRow ->
		          raise (Unify_failure ("Rigid row var cannot be unified with empty closed row\n"))
              | FlexibleRow _ ->
		          Unionfind.change point (RigidRow var)
              | RigidRow var' when var=var' -> ()
              | RigidRow var' ->
		          raise (Unify_failure ("Incompatible rigid row variables cannot be unified\n"))
              | _ -> assert false in

	      let rec extend point =
	        (* point should be a row variable *)
	        let row_var = Unionfind.find point in
	          match row_var with
		        | ClosedRow ->
		            if is_empty_row extension_row then
		              close_empty_row_var extension_row_var
		            else
		              raise (Unify_failure ("Closed row cannot be extended with non-empty row\n"
					                        ^string_of_row extension_row))
		        | RigidRow var ->
		            if is_empty_row extension_row then
		              rigidify_empty_row_var var extension_row_var
		            else
		              raise (Unify_failure ("Rigid row variable cannot be unified with non-empty row\n"
					                        ^string_of_row extension_row))
		        | FlexibleRow var ->
		            if TypeVarSet.mem var (free_row_type_vars extension_row) then
		              raise (Unify_failure ("Unexpected recursive row variable"))
		            else if StringMap.is_empty extension_field_env then
		              match extension_row_var with
			            | point' ->
			                Unionfind.union point point'
		            else
		              Unionfind.change point (BodyRow extension_row)
		        | RecursiveRow _ ->
		            raise (Unify_failure ("Unexpected recursive row variable"))
		        | BodyRow row ->
		            ignore (unify_record_rows uenv (row, extension_row))  (* TODO: check this *)
	      in
	        extend row_var in

      let make_extension_field_env sigma ext_labs ext_fsm =
	    List.fold_left (fun fsm lab -> match StringMap.find lab ext_fsm with
			              | `Absent -> raise Impos
			              | `Present t -> 
			                  let t' = Subst.apply sigma t in
				                StringMap.add lab (`Present t') fsm
		               ) StringMap.empty ext_labs in

      let get_all_name_bindings labs (fsm, _) = 
        List.fold_left (fun names lab -> match StringMap.find lab fsm with
                            `Absent -> names
                          | `Present t -> (match Fable.get_name_bindings t.ftype with
                                               [], None -> names
                                             | [], Some n -> StringSet.add n names
                                             | _ -> raise Impos)
                       ) StringSet.empty labs 
      in
      let get_all_free_names omega labs (fsm, _) = 
        List.fold_left (fun names lab -> match StringMap.find lab fsm with
                            `Absent -> names
                          | `Present t -> 
                              let nlist = Fable.free_names omega t in
                                StringSet.union names (StringSet.from_list nlist)
                       ) StringSet.empty labs 
      in

	    
      (************* Record unification rule for one rigid and one open row
		             
                     sigma_0 = {}            
                     i \in 1..n   p, Om |- ti < sigma_{i-1} (si) : sigma_i'
                     sigma_i = sigma_0, ..., sigma_{i-1}, sigma_i', (yi -> xi)
		             sigma_n* = sigma_n^{-1}   rv = sigma_n*(fn+1:tn+1, ..., fm:tm)
                     ------------------------------------------------------------------------
                     p, Om |- (f1: ~x1:t1, ..., fm :tm) < (f1: ~y1:s1, ..., fn :sn | rv) 
      *************)                                    
      let unify_one_open uenv (rigid_row, open_row) = 
        let fields_are_compatible (lrow, rrow) =
          let llabs = List.map fst (Fable.row_labels_inorder ~filter_absent:true (fst lrow)) in (* List.sort compare (StringSet.elements (get_present_labels lrow)) in *)
          let rlabs = List.map fst (Fable.row_labels_inorder ~filter_absent:true (fst rrow)) in (* List.sort compare (StringSet.elements (get_present_labels rrow)) in *)
	  let extension_labels l r = 
	    let rec drop_lprefix pfx (l,r) = match l,r with
	      | hd1::_, hd2::_ when hd1=hd2 -> pfx, l
	      | hd1::tl, _ -> drop_lprefix (hd1::pfx) (tl, r)
	      | _ -> pfx, l in
	    let rec find_lsuffix (l,r) = match l,r with
	      | _, [] -> l
	      | hl::tl, hr::tr when hl=hr -> find_lsuffix (tl,tr)
	      | _ -> raise (Unify_failure ("Open row\n "^ string_of_row rrow
					   ^"\nand dependent rigid row\n "^ string_of_row lrow
					   ^"\n could not be unified because the former contains labels not in the latter")) in
	    let pfx, l' = drop_lprefix [] (l,r) in
	    let bindings = get_all_name_bindings pfx lrow in
	    let fns = get_all_free_names uenv.omega rlabs rrow in
	      if (StringSet.inter bindings fns) <> StringSet.empty then 
		raise (Unify_failure ("Projection of a field with free names is not permitted; use \"open e as x\" instead"));
	      pfx@(find_lsuffix (l',r))
	  in 
	    (rlabs, extension_labels llabs rlabs) 
	in
	  if not(is_rigid_row rigid_row && not(is_rigid_row open_row)) then
	    failwith "unify_one_open polarity screwed up";
          let (lfield_env', rigid_row_var') as lrow', rigid_rec_row = unwrap_row lrow in
          let (rfield_env', open_row_var') as rrow', open_rec_row = unwrap_row rrow in
          let _ = match rigid_rec_row, open_rec_row with
              None, None -> ()
            | _ -> raise (Unify_failure "Unexpected recursive row variable in a record with fable qualifiers") in
          let (common, extension) = fields_are_compatible (lrow', rrow') in
          let sigma, sigma_inv = unify_field_envs uenv common (lfield_env', rfield_env') in
	      let open_extension = make_extension_field_env sigma extension lfield_env' in
          let _ = unify_row_var_with_row uenv (open_row_var', (open_extension, rigid_row_var')) in
	        sigma  in
	    
      let unify_both_open uenv (lrow, rrow) = 
            let fields_are_compatible (lrow, rrow) =
              let llabs = List.map fst (Fable.row_labels_inorder ~filter_absent:true (fst lrow)) in (* List.sort compare (StringSet.elements (get_present_labels lrow)) in *)
          let rlabs = List.map fst (Fable.row_labels_inorder ~filter_absent:true (fst rrow)) in (* List.sort compare (StringSet.elements (get_present_labels rrow)) in *)
              let rec common_prefix pfx = function
                  h1::t1, h2::t2 ->
                        if h1=h2 then
                          common_prefix (h1::pfx) (t1,t2)
                        else 
                          List.rev pfx, (t1, t2)
                | l1, l2 -> 
                      List.rev pfx, (l1, l2) in
              let pfx, (llabs', rlabs') = common_prefix [] (llabs,rlabs) in
              let bindings = StringSet.union (get_all_name_bindings llabs' lrow) (get_all_name_bindings rlabs' rrow) in
                if bindings <> StringSet.empty then
                  raise (Unify_failure "Unification of open rows only permitted when the missing fields of each don't have name bindings");
                let commonlabs = list_intersect llabs' rlabs' in
                let llabs' = list_difference llabs' commonlabs in
                let rlabs' = list_difference rlabs' commonlabs in
                  (pfx, commonlabs, llabs', rlabs') in

        let (lfield_env', lrow_var') as lrow', lrec_row = unwrap_row lrow in
        let (rfield_env', rrow_var') as rrow', rrec_row = unwrap_row rrow in
        let _ = match lrec_row, rrec_row with
            None, None -> ()
          | _ -> raise (Unify_failure "Unexpected recursive row variable in a record with fable qualifiers") in
	    let (pfx, common, llabs, rlabs) = fields_are_compatible (lrow, rrow) in
	    let sigma, sigma_inv = unify_field_envs uenv (pfx@common) (lfield_env', rfield_env') in
	    let fv = fresh_row_variable() in
	      if (Unionfind.equivalent lrow_var' rrow_var') then
		match llabs, rlabs with (* row vars are the same; so, cannot extend each other *)
		    [], [] -> 
		      let _ = unify_row_var_with_row uenv (lrow_var', (StringMap.empty, fv)) in
			sigma
		  |  _ -> raise (Unify_failure "Unification of identical open rows with unequal extensions")
	      else
		let extends_r = make_extension_field_env sigma llabs lfield_env' in
		let extends_l = make_extension_field_env sigma_inv rlabs rfield_env' in
		  (* let examine_row_var pfx rv = match Unionfind.find rv with *)
		  (* 	      | ClosedRow -> print_string (pfx ^ "Closed row var\n") *)
		  (* 	      | FlexibleRow i -> print_string (pfx ^ "Flexible row var" ^ (string_of_int i) ^ "\n") *)
		  (* 	      | RigidRow i -> print_string (pfx ^ "Rigid row var" ^ (string_of_int i) ^ "\n") *)
		  (* 	      | RecursiveRow (i, tdr) -> print_string (pfx ^ "Recursive row var" ^ (string_of_int i) ^ "\n") *)
		  (* 	      | BodyRow tdr ->  print_string (pfx ^ "Body row var\n") in *)
		  (*  	      examine_row_var "lrow_var: " lrow_var'; 	       *)
		  (* 	      examine_row_var "rrow_var: " rrow_var'; *)
		  (* 	      examine_row_var "fresh_row_var: " fv; *)
		let _ = unify_row_var_with_row uenv (lrow_var', (extends_l, fv)) in
 		let _ = unify_row_var_with_row uenv (rrow_var', (extends_r, fv)) in
		  sigma
      in
	
	    if is_rigid_row lrow then
	      if is_rigid_row rrow then
            unify_both_rigid uenv (lrow, rrow)
	      else
	        unify_one_open uenv (lrow, rrow)
	    else
	      if is_rigid_row rrow then 
	        unify_one_open (flip_polarity uenv) (rrow, lrow)
	      else
	        unify_both_open uenv (lrow, rrow)
    in
      
    let is_empty_fquals row = 
      let rec check_empty_fquals rec_vars (fsm, rv) = 
        (StringMap.iter 
           (fun _  -> function
		      | `Absent -> ()
		      | `Present t -> 
                  if not (Fable.is_empty t.ftype) then 
                    raise (Unify_failure "Unification of non-rigid rows with fable qualifiers; not yet supported"))
           fsm;
         match Unionfind.find rv with 
           | RecursiveRow (v, r) when not (IntMap.mem v rec_vars) -> 
		       check_empty_fquals (IntMap.add v r rec_vars) r
           | BodyRow r -> 
		       check_empty_fquals rec_vars r
           | _ -> ()) in
	    try
	      let _ = check_empty_fquals IntMap.empty row in
	        true
	    with 
	        Unify_failure _ -> false in
      match uenv.polarity with 
	      `Neg -> unify_dependent_rows (flip_polarity uenv) (rrow, lrow)
	    | `Pos -> unify_dependent_rows uenv (lrow, rrow)
            
  let initial_uenv phantoms omega exp_opt sub = 
    {rec_types = IntMap.empty;
     rec_rows = IntMap.empty;
     omega = omega;
     phantoms = phantoms;
     exp = exp_opt;
     polarity = `Pos;
     subtype = sub} 
      
  let unify phantoms omega exp_opt sub (t1, t2) =
    let uenv = initial_uenv phantoms omega exp_opt sub  in
      unify' uenv (t1, t2) 
        
  (* Debug.if_set (show_unification) (fun () -> "Unified types: " ^ string_of_datatype t1) *)
  and unify_rows omega sub (row1, row2) =
    let uenv = initial_uenv [] omega None sub in
      unify_variant_rows uenv (row1, row2)

  let unify_old omega ts = ignore (unify [] omega None false ts)
    
end

open Unify

exception UndefinedVariable of string
exception Bad_fable_type of string
exception Bad_type_annotation of Types.datatype * string

let norm_ultype :  Types.u_ldatatype -> ldatatype = 
  fun lt -> let {ltype=lt'; ftype=_} = Sugar.normalize_type {ltype=lt;ftype=`Empty} in lt'
                                                                                         
let db_descriptor_type =
  snd (Parse.parse_string (Parse.checked_datatype Sugar.normalize_type) "(driver:String, name:String, args:String)")

(* extract data from inference_expressions *)
let type_of_expression : expression -> datatype =
  fun exp -> let `T (_, t, _) = expression_data exp in t
let pos_of_expression : expression -> position =
  fun exp -> let `T (pos, _, _) = expression_data exp in pos

(* type typing_environment = environment * alias_environment *)

let rec extract_row_opt : Typeenv.omega -> datatype -> row option = fun omega (* ((env, alias_env) as typing_env)  *) t ->
  match t.ltype with
    | `Record row -> Some row
    | `Variant row -> Some row
    | `MetaTypeVar point ->
        begin
          match Unionfind.find point with
            | Body t -> extract_row_opt omega t
	        | Recursive (i, t) -> extract_row_opt omega t (* The way type annotations are handled means that we must include this case *)
            | _ -> None 
        end
    | `Application (s, ts) ->
        let vars, alias = Typeenv.lookup_alias (s, ts) omega in 
          extract_row_opt omega (instantiate_alias (vars, alias) ts)
    | _ -> None

let extract_row omega dt = 
  match extract_row_opt omega dt with 
      None -> failwith
        ("Internal error: attempt to extract a row from a datatype that is not a record or variant: " ^ (string_of_datatype dt))
    | Some r -> r

let instantiate_assumption : Typeenv.omega -> (quantifier list * datatype) -> datatype =
  fun omega (quantifiers, t) -> 
    if quantifiers = [] then
      Subst.rename_bound_vars t
    else
      let _ = Debug.if_set (show_instantiation)
        (fun () -> "Instantiating assumption: " ^ (string_of_assumption (quantifiers, t))) in
      let tenv, renv = List.fold_left
        (fun (tenv, renv) (var, k) -> 
           let annot_kind dt = 
	     let kv = Fable.fresh_kind_variable k in 
	       {ltype = dt.ltype; ftype = `KindVar kv} in  (* preserve tvar kind *)
             match var with
               | `TypeVar var -> IntMap.add var (annot_kind (fresh_type_variable ())) tenv, renv
               | `RigidTypeVar var -> IntMap.add var (annot_kind (fresh_type_variable ())) tenv, renv
                 | `RowVar var -> tenv, IntMap.add var (fresh_row_variable ()) renv
        ) (IntMap.empty, IntMap.empty) quantifiers
      in
        Subst.rename_bound_vars (instantiate_datatype (tenv, renv) t)
	  
(** instantiate env var
    Get the type of `var' from the environment, and rename bound typevars.
*)
let instantiate : Typeenv.omega -> string -> datatype = fun omega var ->
  try
    let quantifiers, t = Typeenv.lookup var omega in
      instantiate_assumption omega (quantifiers, t)
  with Not_found ->
    raise (UndefinedVariable ("Variable '"^ var ^"' does not refer to a declaration"))

(*
  get the quantifiers for a datatype
  i.e. the quantifiers required to close a datatype
  
  (the first argument specifies type variables that should remain free)
*)
let rec get_quantifiers : TypeVarSet.t -> datatype -> quantifier list = 
  fun bound_vars dt -> match dt.ltype with 
    | `Not_typed -> failwith "Internal error: Not_typed encountered in get_quantifiers"
    | `Primitive _ -> []
    | `MetaTypeVar point ->
        (match Unionfind.find point with
           | Flexible var
           | Rigid var when TypeVarSet.mem var bound_vars -> []
           | Flexible var -> [`TypeVar var, Fable.kind_qual dt.ftype]
           | Rigid var -> [`RigidTypeVar var, Fable.kind_qual dt.ftype]
           | Recursive (var, body) ->
               Debug.if_set (show_recursion) (fun () -> "rec (get_quantifiers): " ^(string_of_int var));
               if TypeVarSet.mem var bound_vars then
                 []
               else
                 get_quantifiers (TypeVarSet.add var bound_vars) body
           | Body t -> get_quantifiers bound_vars t)
    | `Function (f, m, t) ->
        let from_gens = get_quantifiers bound_vars f
        and mailbox_gens = get_quantifiers bound_vars m
        and to_gens = get_quantifiers bound_vars t in
          unduplicate (=) (from_gens @ mailbox_gens @ to_gens)
    | `Record row
    | `Variant row -> get_row_quantifiers bound_vars row 
    | `Table (r, w) -> unduplicate (=) (get_quantifiers bound_vars r @ get_quantifiers bound_vars w)
    | `Application (_, args) ->
        unduplicate (=) (Utility.concat_map (get_quantifiers bound_vars) args)

and get_row_var_quantifiers : TypeVarSet.t -> row_var -> quantifier list =
  fun bound_vars row_var ->
    match Unionfind.find row_var with
      | ClosedRow -> []
      | FlexibleRow var
      | RigidRow var when TypeVarSet.mem var bound_vars -> []
      | FlexibleRow var
      | RigidRow var -> [`RowVar var, FableType.UKind] (* TODO : row var kinds? *)
      | RecursiveRow (var, rec_row) ->
          Debug.if_set (show_recursion) (fun () -> "rec (get_row_var_quantifiers): " ^(string_of_int var));
          (if TypeVarSet.mem var bound_vars then
             []
           else
             get_row_quantifiers (TypeVarSet.add var bound_vars) rec_row)
      | BodyRow row -> get_row_quantifiers bound_vars row

and get_field_spec_quantifiers : TypeVarSet.t -> field_spec -> quantifier list =
  fun bound_vars ->
    function
      | `Present t -> get_quantifiers bound_vars t
      | `Absent -> []

and get_row_quantifiers : TypeVarSet.t -> row -> quantifier list =
  fun bound_vars (field_env, row_var) ->
    let field_vars = StringMap.fold
      (fun _ field_spec vars ->
         get_field_spec_quantifiers bound_vars field_spec @ vars
      ) field_env [] in
    let row_vars = get_row_var_quantifiers bound_vars (row_var:row_var)
    in
      field_vars @ row_vars

let type_mismatch ~expected ~inferred ~pos ~src msg =
  raise (Type_error (pos, 
                     src ^" has type "^string_of_datatype inferred
                     ^" but is annotated with type "
                     ^ string_of_datatype expected^"\n"^
                       msg))

(** generalise: 
    Universally quantify any free type variables in the expression.
*)
let generalise : Typeenv.omega -> datatype -> assumption = 
  fun omega t ->
    let vars_in_env = Typeenv.all_type_vars omega in 
    let quantifiers = get_quantifiers vars_in_env t in
      Debug.if_set (show_generalisation) (fun () -> "Generalised: " ^ (string_of_assumption (quantifiers, t)));
      (quantifiers, t) 


let annot_ukind lt =
  {ltype=lt; ftype= `Kind UKind}  

let annot_mkind lt =
  {ltype=lt; ftype= `Kind MKind}  

let annot_kindvar lt k =
  let kv = Fable.fresh_kind_variable k in 
    {ltype=lt; ftype= `KindVar kv}

let rec constant_type = 
  let constant_ltype = function
    | Boolean _ -> `Primitive `Bool
    | Integer _ -> `Primitive `Int
    | Float _ -> `Primitive `Float
    | Char _ -> `Primitive `Char in
    function
        String _ -> string_type ()
      | t -> annot_ukind (constant_ltype t)

let expand_record_open exp var_types = match exp with
  | Record_unpack(fields, value, body, None, _) -> 
      let tmp = Utility.gensym ~prefix:"open" () in
      let value_var = Variable(tmp, `T(dummy_position, type_of_expression value, None)) in
      let body = List.fold_left2 (fun body (fieldname, var) (var', t) -> 
				                    if (var <> var') then
				                      raise Impos;
				                    Record_selection(fieldname, var, Utility.gensym ~prefix:"ignore" (),
						                             value_var, body, 
						                             `T(dummy_position, type_of_expression body, None))
				                 ) body fields var_types in
	    Let (tmp, value, body, `T(dummy_position, type_of_expression body, None))
  | _ -> raise Impos



let infer_label_type omega = 
  let rec infer_label_type rec_vars t = 
    let join_status = function
	_, `Not_label
      | `Not_label, _ -> `Not_label
      | _, `General 
      | `General, _ ->  `General
      | _ -> `Specific in
      try
	match Fable.get_label_type_expr t.ftype with 
            None -> `General
          | Some e -> `Specific
      with 
          Not_found -> 
            let t = concrete_type t in 
              match t.ltype with (* if possible, try to give this a label qualifier *)
                  _unit when (_unit = (Types.unit_type()).ltype) -> 
                    `Specific
		      
		| `Record row when is_closed_row row ->
                    StringMap.fold (fun _ t status -> 
                                      match t, status with 
                                          `Absent, _
					|  _, `Not_label -> status
					| `Present t, _  -> 
                                            join_status (status, infer_label_type rec_vars t)
                                   ) (fst row) `Specific 

		| `Variant row when is_closed_row row -> 
                    StringMap.fold (fun label spec status -> match spec with
					`Absent -> status
                                      | `Present t -> join_status (status, infer_label_type rec_vars t)
                                   ) (fst row) `Specific
                      
		| `Application (name, ts) -> 
                    let vars, def = Typeenv.lookup_alias (name, ts) omega in
                    let inst = Instantiate.instantiate_alias (vars, def) ts in
                      infer_label_type rec_vars inst

		| `MetaTypeVar pt -> 
                    begin
                      match Unionfind.find pt with 
                          Recursive (ix, body) -> 
                            if IntSet.mem ix rec_vars then 
                              `General
                            else
                              infer_label_type (IntSet.add ix rec_vars) body
			| _ -> `Not_label
                    end

		| _ -> `Not_label in
    infer_label_type IntSet.empty
      
(* let check_label_type env : 'a Types.ldatatype' -> bool =  *)
(*   fun lt ->  *)
(*     let t = {ltype=lt;ftype=`Empty} in *)
(*       match infer_label_type env t with  *)
(*           `Not_label -> raise (Bad_fable_type ("Type "^string_of_datatype t^ *)
(*                                                  " cannot be used as a label")) *)
(*         | _ -> true *)

let check_label_type env : 'a Types.ldatatype' -> bool = 
  fun lt -> true

let check_no_escaping_names : Typeenv.omega -> Types.datatype -> unit = 
  fun omega d -> match Fable.free_names ~expect_empty:true omega d with
      [] -> ()
    | [name] -> 
	    raise (Bad_fable_type ("The following name escapes its scope : " ^name))
    | l -> 
	    let names = List.fold_right (fun name s -> (name ^ ", " ^s)) l "" in
	      raise (Bad_fable_type ("The following names escape their scope : " ^names))

let rec check_type : Typeenv.omega -> Types.u_datatype -> Types.datatype * FableType.kind =

  let strip_name_bindings_utype {ltype=lt;ftype=fq} = {ltype=lt; ftype = Fable.strip_name_bindings fq} in

  let compatible_label_constraint : Typeenv.omega -> expression -> datatype -> unit =
    fun env exp lt ->
      (* TODO: What about rigid vs open rows here? Probably want them both to be rigid *)
      ignore (Unify.unify [] env None false (Syntax.node_datatype exp, lt)) in
    
  let rec check_fq env (lt : ldatatype) fq =
    ignore (FableType.well_formed_fq fq);
    let env = Typeenv.clear_type_annotation env in
      match fq with
	| `Empty -> `Kind UKind, UKind
	    
	| `Kind k -> `Kind k, k
	    
	| `Label(eopt, fq') ->
            let _, k' = check_fq env lt fq' in (* TODO: ensure that k' is unifiable with UKind if lt is a type variable *)
            let _ = (k'=UKind) in (*  || check_label_type env lt) in  *)
              begin
		match eopt with
                    `Nothing -> `Label(`Nothing, `Kind UKind),  UKind
                  | `Required e ->
                      let e' = type_check_expression env e in 
                      let _ = compatible_label_constraint env e' (def_fable_annot lt) in
                      `Label(`Required e', `Kind UKind), UKind
		  | `Inferred pt -> raise Impos (* should never occur in an unchecked fable type *)
              end
		
	| `Labeling(e, fq') ->
            let fq'', _ = check_fq env lt fq' in
            let e' = type_check_expression env e in
            let _ =  (Syntax.node_datatype ->- (fun x-> (Fable.kind_qual x.ftype) = UKind)) e' in (*  = UKind || check_label_type env x.ltype))) e' in *)
              `Labeling(e', fq''), MKind
		
      | _ -> raise Impos in (* No name bindings allowed at this level *)
    
  let datatype_kind {ltype=_;ftype=fq} = Fable.kind_qual fq in    

  let rec check_ultype env : u_ldatatype -> ldatatype * kind option = function
    | `Primitive p -> `Primitive p, Some UKind

    | `MetaTypeVar pt ->
        begin
          match Unionfind.find pt with  (* Ensure that the kind here is consistent with the kind in the uft arg *)
              Flexible _

            | Rigid _ -> `MetaTypeVar pt, None (* kind not yet determined; to be fixed during unification *)
		        
	        | UncheckedRecursive (var, t) -> 
		        (* Produced during desugaring, without having first being checked *)
		        if (Typeenv.contains_rec_var var env) then 
		          `MetaTypeVar pt, None 
		        else 
		          let env' = Typeenv.add_rec_var var env in 
		          let t', k = check_type env' t in
		            Unionfind.change pt (Recursive(var, t'));
		            `MetaTypeVar pt, Some k
		              
	        | Recursive (_, t) (* tv has already been unified with a checked datatype. no check necessary *)		
            | Body t -> 
		        `MetaTypeVar pt, Some (datatype_kind t)
		          
        end
          
    | `Function(arg, mb, ret) ->
        let (phantoms, name) = Fable.get_name_bindings arg.ftype in
          assert_msg (List.for_all (not -<- (flip Typeenv.contains) env) phantoms) "Duplicate phantom name";
          assert_msg (match name with None -> true | Some n -> not (Typeenv.contains n env)) "Duplicate name";
          let env' = Fable.add_phantom_bindings phantoms env in
          let arg = strip_name_bindings_utype arg in
            (* check FV(t)\dom(Omega) = phantoms *)
          let fv_arg = Fable.free_names env arg in
            assert_msg (List.for_all (fun n -> List.mem n phantoms) fv_arg) "Phantom name omitted";
            assert_msg (List.for_all (fun n -> List.mem n fv_arg) phantoms) "Extra phantom name";
            let arg', _ = check_type env' arg in
            let env'' = match name with
                None -> env'
              | Some name -> Typeenv.add_normal_binding name ([], arg') env in (* Q: is it right to generalize all fvs here? A: No! *)
            let ret',_ = check_type env'' ret in
            let mb',_ = check_type env'' mb in (* TODO: revise mailbox typing *)
              `Function(Fable.restore_bindings phantoms name arg', mb', ret'), Some UKind

    | `Record ((fsm, rv) as row) ->
        (*         let row = flatten_row row in  *)
        let fnames = List.map fst (Fable.row_labels_inorder (fst row)) in 
          (* List.sort compare (StringMap.fold (fun key _ l -> key::l) fsm []) in (\* name bindings scopes are from L-to-R in field name order *\) *)
        let check_one_field (env, fsm') fn =
          match StringMap.find fn fsm with
              `Present ({ltype=lt; ftype=fq}) ->
                begin
                  match Fable.get_name_bindings fq with
                      [], n ->
                        let lt',_ = check_ultype env lt in
                        let fq = Fable.strip_name_bindings fq in
                        let fq', k = check_fq env lt' fq in
                        let field_type' = {ltype=lt'; ftype = fq'} in
                        let env' = match n with
                            None -> env
                          | Some name -> Typeenv.add_normal_binding name ([], field_type') env in
                        let field_type' = Fable.restore_bindings [] n field_type' in
                          env', StringMap.add fn (`Present field_type') fsm' 
                    | _ -> raise (Bad_fable_type "Unexpected phantom bindings in a record type")
                end
            | `Absent -> env, StringMap.add fn `Absent fsm' in 
        let _, fsm' = List.fold_left check_one_field (env, StringMap.empty) fnames in
          `Record(fsm',rv), Some UKind
	        
    | `Variant ((fsm, rv) as row) ->
	    let fsm' = StringMap.fold (fun key fs fsm' -> match fs with
				                       `Absent -> StringMap.add key `Absent fsm'
				                     | `Present t ->
					                     let t', k = check_type env t in
					                       StringMap.add key (`Present t') fsm'
				                  ) fsm StringMap.empty in
	    let _ = match Unionfind.find rv with
	        ClosedRow
	      | FlexibleRow _
	      | RigidRow _
	      | BodyRow _ -> ()
	      | RecursiveRow _ -> raise (Bad_fable_type "Recursive row variables are not supported") in
	      `Variant(fsm', rv), Some UKind (* Can this be an LKind? *)

    | `Application(alias, args) -> 
	    `Application(alias, List.map (fst -<- (check_type env)) args), Some UKind
	      
    | ult -> norm_ultype ult, Some UKind (* failwith ("check_ultype not yet fully implemented ")  *)in
    
    fun env ({ltype=lt; ftype=fq} as t) -> 	
      let env = Typeenv.set_phantom_phase env in
        ignore (FableType.well_formed_fq fq);
        match Fable.get_name_bindings fq with
	        [], None ->
	          let lt', kind' = check_ultype env lt in
	          let fq', kind = check_fq env lt' fq in
		        {ltype=lt'; ftype=fq'}, kind
	      | _ -> raise (Bad_fable_type "Unexpected top-level name bindings")
	          
and arg_var_mapping omega vlist : datatype option -> (omega * expression list * datatype * datatype option * FableType.fable_type option) = function
    None -> (* no top-level type annotation; but the annots may be inlined in the vars *)
      let (omega, vlist', vtypes) = 
	    List.fold_left (fun (omega, vlist, vtypes) -> function
			                Variable(name,`U pos) -> 
			                  let t = fresh_type_variable () in
			                  let omega' = Typeenv.add_normal_binding name ([], t) omega in
			                  let v' = Variable(name, `T(pos, t, None)) in
				                (omega', v'::vlist, t::vtypes)
			              | HasType(Variable(name,`U pos), t, _) -> 
			                  let t, _ = check_type omega t in
			                  let omega' = Typeenv.add_normal_binding name ([], t) omega in
			                  let v' = Variable(name, `T(pos, t, None)) in
				                (omega', v'::vlist, t::vtypes)
			              | _ -> failwith "Bad desugaring of `FunLit variables")
	      (omega, [], []) vlist in
	    omega, rev vlist', make_tuple_type (List.rev vtypes), None, None
	      
  | Some t -> (* No inline annotations permitted *)
      let omega = match Fable.get_name_bindings t.ftype with
          [], _ -> omega 
        | phantoms, _ -> Fable.add_phantom_bindings phantoms omega in
      let t = concrete_type t in
        match t.ltype with  
            `Function({ltype =`Record ((fsm, _) as row); ftype = _} as arg, _, return_type) when is_closed_row row -> 
              let fields = 
                StringMap.fold (fun fieldname t accu -> match t with
                                    `Present t -> (fieldname, t) :: accu
                                  | `Absent -> raise (Bad_fable_type "Fable type annotations are only permitted on positive records")
                               ) fsm [] in
              let fields = rev fields in
                if (length vlist <> length fields) then
                  raise (Bad_fable_type "Ascribed signature requires a different number of fields");
                let (omega, vlist', refinement_fsm) =
                  List.fold_left2 (fun (omega, accu, fsm) (field_name, field_type) -> function
                                       Variable (name, `U pos) -> 
                                         let fsm' = StringMap.add field_name (Variable(name, `T(pos, field_type, None))) fsm in
                                         let _, nopt = Fable.get_name_bindings field_type.ftype in
                                         let field_type', omega' = match nopt with  (* this could be just a warning; but it's most likely an error *)
                                             Some n when n <> name -> raise (Bad_fable_type ("Term/type naming inconsitency : " ^ n ^ " <> " ^name))
                                           | Some n -> 
                                               let field_type = Fable.strip_name_bindings_type field_type in
                                               let field_type = Subst.rename_bound_vars field_type in
                                                 (field_type,
                                                  Typeenv.add_noshadow_binding name ([], field_type) omega)
                                           | _ -> 
                                               let field_type = Subst.rename_bound_vars field_type in
                                                 (field_type, 
                                                  Typeenv.add_normal_binding name ([], field_type) omega) in
                                         let vexp' =  Variable(name, `T(pos, field_type', None)) in
                                           (omega', vexp'::accu, fsm')
                                     | HasType _ -> raise (Bad_fable_type "Cannot mix type signatures with inline type ascriptions")
                                  ) (omega, [], StringMap.empty) fields vlist in
                let omega =
                  match Fable.get_name_bindings arg.ftype with
                      phantoms, None -> 
                        Fable.add_phantom_bindings phantoms omega
                    | phantoms, Some n -> 
                        let arg' = Fable.strip_name_bindings_type arg in
                        let omega' = Typeenv.add_noshadow_binding n ([],arg') omega in
                        let omega' = Fable.add_phantom_bindings phantoms omega' in 
                        let namevar = Variable(n, `T(dummy_position, t, None)) in
                          StringMap.fold (fun tupleindex (Variable(_, data) as arg) omega -> 
                                            let refinement = (arg, Project(namevar, tupleindex, data)) in
                                              Typeenv.add_refinement refinement omega
                                         ) refinement_fsm omega'
                    | _ -> omega in
                  (omega, rev vlist', arg, Some return_type, Some t.ftype)
          | _ -> raise (Bad_fable_type ("Cannot yet handle ascribed type of the form: " ^ (DerivingSyntax.Show_datatype.show t)))

	      
and type_check_expression : Typeenv.omega -> Syntax.untyped_expression -> expression =
  let aux omega expression =
    let fannot lt = 
      let t = Types.def_fable_annot lt in
      let t', _ = check_type omega t in
	t'
    and unify = Unify.unify [] omega None false
    and unify_rows = Unify.unify_rows omega false
    in
      try
        Debug.if_set (show_typechecking) (fun () -> "Typechecking expression: " ^ (string_of_expression expression));
        match (expression : Syntax.untyped_expression) with
          | Constant (value, `U pos) -> Constant (value, `T (pos, constant_type value, None))

          | Variable (name, `U pos) ->
	      if ((name="unlabel" || name="relabel") && not (Typeenv.check_context omega `Policy)) then
		raise (Bad_fable_type ("Illegal usage of " ^name^ " in non-policy code"))
	      else
		let t = instantiate omega name in
		let label = match lookup_location name omega with
		  | `Policy -> Some "policy"
		  | _ -> None in
		  Variable (name, `T (pos, t, label))
		    
          | Abs (f, `U pos) -> 
	      (* TODO: hmmm ... can't understand why this construct is of any use. *)
	      (* links> var z = abs fun (x) { x };; *)
	      (* z = abs fun : *(|a) -> (|a) *)
	      (* links> z (5);; *)
	      (* (1=5) : (1:Int) *)
	      let omega = Typeenv.clear_type_annotation omega in
              let f = type_check_expression omega f in
              let arg_type = fannot (`Record (make_empty_open_row ())) in
              let result_type = fresh_type_variable () in
              let mb_type = fresh_type_variable () in
              let f_ltype = `Function (make_tuple_type [arg_type], 
                                       mb_type, 
                                       result_type) in
                unify (type_of_expression f, annot_ukind f_ltype); (* if we're applying f, then it had better not be labeled *)
                let etype =  `Function (arg_type, mb_type, result_type) in
                  Abs (f, `T (pos, annot_ukind etype, None))

          | App (f, p, `U pos) -> (* Question: Again ... not sure how this differs from Apply or what its purpose is *)
	      let omega = Typeenv.clear_type_annotation omega in
              let f = type_check_expression omega f
              and p = type_check_expression omega p 
              and result_type = fresh_type_variable ()
              and mb_type = instantiate omega "_MAILBOX_" 
              in
                (* not really necessary, but might catch some errors *) 
                unify (type_of_expression p, def_fable_annot (`Record (make_empty_open_row ()))); 
                unify (type_of_expression f,
                       annot_ukind (`Function (type_of_expression p, mb_type, result_type)));
                App (f, p, `T (pos, result_type, None))

          | Apply (f, ps, `U pos) ->
	      let omega = Typeenv.clear_type_annotation omega in
              let f = type_check_expression omega f in
              let ps,ps' = match f, ps with 
		  Variable("relabel", _), [x; label] -> 
		    let x' = type_check_expression omega x in
		    let label' = type_check_expression (Typeenv.set_phantom_phase omega) label in
		      ([x'; label'], 
		       [x'; unit_expression (Syntax.expression_data label')]) (* hacky rewrite to avoid passing phantom labels at runtime *)
		| _ -> 
		    let ps = List.map (type_check_expression omega) ps in
		      ps, ps in
              let mb_type = instantiate omega "_MAILBOX_" in
              let f_type = concrete_type (type_of_expression f) in
              let arg_type = make_tuple_type (List.map type_of_expression ps) in
	      let return_type =
                try 
		  match f_type.ltype with
		      `Function (formal, formal_mb, ret) -> 
			if Fable.kind_qual f_type.ftype <> UKind then
			  raise (Bad_fable_type "Attempt to apply a labeled function")
			    (* 			      mistyped_application pos (f, f_type) (ps, List.map type_of_expression ps) (Some mb_type) *)
			else
			  let phantoms, nopt = Fable.get_name_bindings formal.ftype in
			  let formal = Fable.strip_name_bindings_type formal in
			  let arg = Syntax.tuple_expr (`T(pos, arg_type, None)) ps in
			  let sigma = Unify.unify phantoms omega (Some arg) true (arg_type, formal) in
			  let sigma = match nopt with 
			      None -> sigma
			    | Some n -> Subst.extend (n, arg) sigma in
			  let formal_mb = Subst.apply sigma formal_mb in
			  let ret = Subst.apply sigma ret in
			    unify (mb_type, formal_mb);
			    ret
		    | _ -> (* must infer f's type... which means it can't have any fable annots
			      so just unify with the expected type *)
			let return_type = fresh_type_variable () in 
			  unify (annot_ukind (`Function(arg_type, mb_type, return_type)), f_type);
			  return_type
                with Unify_failure _ ->
                  mistyped_application pos (f, f_type) (ps, List.map type_of_expression ps) (Some mb_type)
	      in
		Apply (f, ps', `T (pos, return_type, None))

          | Condition (if_, then_, else_, `U pos) ->
              let if_ = type_check_expression (Typeenv.clear_type_annotation omega) if_ in
              let _ = 
                try unify (type_of_expression if_, bool_type ())
                with Unify_failure _ -> 
                  mistype
                    (pos_of_expression if_)
                    (if_, type_of_expression if_)
                    (fannot (`Primitive `Bool)) in
              let then_ = type_check_expression omega then_ in
              let else_ = type_check_expression omega else_ in
              let _ = unify (type_of_expression then_, type_of_expression else_) in
	      let result_type = type_of_expression else_ in
                Condition (if_, then_, else_, `T (pos, result_type, None)) 
		  
          | Comparison (l, oper, r, `U pos) ->
	      let omega = Typeenv.clear_type_annotation omega in
              let l = type_check_expression omega l in
              let r = type_check_expression omega r in
                unify (type_of_expression l, type_of_expression r);
                Comparison (l, oper, r, `T (pos, bool_type (), None))
		  
          | Abstr (variables, body, `U pos) ->
              let mb_type = fresh_type_variable () in
              let (body_omega, variables', arg_type, rettype_opt, ftype_opt) = arg_var_mapping omega variables (Typeenv.get_type_annotation omega) in
              let body_omega = Typeenv.add_normal_binding "_MAILBOX_" ([], mb_type) body_omega in
              let body = type_check_expression (Typeenv.set_type_annotation_opt body_omega rettype_opt) body in
              let type' = `Function (arg_type, mb_type, type_of_expression body) in
	      let typ = match ftype_opt with
		  Some fq -> {ltype = type'; ftype= fq}
		| _ -> annot_ukind type' in
                Abstr (variables', body, `T (pos, typ, None))
                  
          | Let (variable, value, body, `U pos) ->
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in
              let vtype = (if is_value value then (generalise omega (type_of_expression value))
                           else ([], type_of_expression value)) in
              let omega' = Typeenv.add_normal_binding variable vtype omega in
              let body = type_check_expression omega' body in        
		check_no_escaping_names omega (type_of_expression body);
                Let (variable, value, body, `T (pos, type_of_expression body, None))

          | Rec (variables, body, `U pos) ->
	      let annot = Typeenv.get_type_annotation omega in
              let best_omega, vars = type_check_mutually (Typeenv.clear_type_annotation omega) variables in
              let body = type_check_expression (Typeenv.set_type_annotation_opt best_omega annot) body in
                Rec (vars, body, `T (pos, type_of_expression body, None))

          | Xml_node (tag, atts, cs, `U pos) -> 
              let separate = partition (is_special -<- fst) in
              let (special_attrs, nonspecial_attrs) = separate atts in
                (* "event" is always in scope for the event handlers *)
              let ev_type = fannot (`Application ("Event", [])) in
              let attr_omega = Typeenv.add_normal_binding "event"  ([], ev_type) omega in
                (* extend the env with each l:name bound variable *)
              let attr_omega = Typeenv.add_normal_binding "_MAILBOX_" ([], fresh_type_variable ()) attr_omega in
              let special_attrs = map (fun (name, expr) -> (name, type_check_expression attr_omega expr)) special_attrs in
                (* Check that the bound expressions have type 
                   <strike>XML</strike> unit. *)
              let contents = map (type_check_expression omega) cs in
	      let omega = Typeenv.clear_type_annotation omega in
              let nonspecial_attrs = map (fun (k,v) -> k, type_check_expression omega v) nonspecial_attrs in
              let attr_type = string_type () in
                (* force contents to be XML, attrs to be strings *)
              let _ = List.iter (fun node -> ignore (unify (type_of_expression node, xml_type ()))) contents in
              let _ = List.iter (fun (_, node) -> ignore (unify (type_of_expression node, attr_type))) nonspecial_attrs in
              let trimmed_node =
                Xml_node (tag, 
                          nonspecial_attrs,         (* +--> up here I mean *)
                          contents,                 (* | *)
                          `T (pos, xml_type (), None))
              in                                    (* | *)
                (* could just tack these on up there --^ *)
                add_attrs special_attrs trimmed_node

          | Record_intro (bs, r, `U pos) ->
	      let omega = Typeenv.clear_type_annotation omega in
              let bs, field_env, absent_field_env =
                StringMap.fold 
                  (fun label e (bs, field_env, absent_field_env)  ->
                     let e = type_check_expression omega e in
                     let t = type_of_expression e in
                       (StringMap.add label e bs,
                        StringMap.add label (`Present t) field_env,
                        StringMap.add label `Absent absent_field_env))
                  bs (StringMap.empty, StringMap.empty, StringMap.empty) in
              let dresult_type = annot_ukind (`Record (field_env, Unionfind.fresh ClosedRow)) in
                begin
                  match r with
                    | None ->
                        Record_intro (bs, None, `T (pos, dresult_type, None))
                    | Some r ->
                        let r = type_check_expression omega r in
                        let rtype = type_of_expression r in
                          (* make sure rtype is a record type! *)
                          
                          unify(rtype, annot_ukind ( `Record (absent_field_env, fresh_row_variable())));
                          
                          let (rfield_env, rrow_var), _ = unwrap_row (extract_row omega rtype) in
                            
                          (* attempt to extend field_env with the labels from rfield_env
                             i.e. all the labels belonging to the record r
                          *)
                          let field_env' =
                            StringMap.fold 
                              (fun label t field_env' ->
                                 match t with
                                   | `Absent ->
                                       if StringMap.mem label field_env then
                                         field_env'
                                       else
                                         StringMap.add label `Absent field_env'
                                   | `Present _ ->
                                       if StringMap.mem label field_env then
                                         failwith ("Could not extend record "^string_of_expression r^" (of type "^
                                                     string_of_datatype rtype^") with "^
                                                     string_of_expression
                                                     (Record_intro (bs, None, `T (pos, dresult_type, None)))^
                                                     " (of type"^string_of_datatype dresult_type ^
                                                     ") because the labels overlap")
                                       else
                                         StringMap.add label t field_env') rfield_env field_env in
                            Record_intro (bs, Some r, `T (pos, annot_ukind (`Record (field_env', rrow_var)), None))                  
                end           
                  
          | Record_selection (label, label_variable, variable, value, body, `U pos) ->
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in
              let label_variable_type = annot_kindvar (fresh_type_variable ()).ltype MKind in (* can only project out of an unlabeled record *)
                unify (type_of_expression value, annot_ukind (`Record (make_singleton_open_row (label, `Present (label_variable_type)))));
                let value_row = extract_row omega (type_of_expression value) in
		  (match Fable.free_names omega label_variable_type with 
		     | [] -> ()
		     | hd::tl -> raise (Bad_fable_type "Illegal projection from a dependent product; use open e1 as e2 instead"));
                  let body_omega = Typeenv.add_normal_binding label_variable ([], label_variable_type) omega in
		  let variable_type = annot_ukind (`Record (row_with (label, `Absent) value_row)) in
                  let body_omega = Typeenv.add_normal_binding variable ([], variable_type) body_omega in
		  let refinement = 
		    let fsm = StringMap.add label (Variable(label_variable, `T(pos, label_variable_type, None))) StringMap.empty in
		      Record_intro(fsm, Some (Variable(variable, `T(pos, variable_type, None))), `T(pos, (type_of_expression value), None)) in
		  let body_omega = Typeenv.add_refinement (value, refinement) body_omega in
                  let body = type_check_expression body_omega body in
                  let body_type = type_of_expression body in
		    check_no_escaping_names omega body_type;
                    Record_selection (label, label_variable, variable, value, body, `T (pos, body_type, None))
		      
	  | Record_unpack (fields, value, body, None, `U pos) -> (* open a dependent product *)
	      let value = type_check_expression (Typeenv.clear_type_annotation omega) value in 
	      let body, ftypes, fields = 
		let value_type = concrete_type (type_of_expression value) in 
		  if not (Fable.sub_kind ((Fable.kind_qual value_type.ftype), UKind)) then 
		    raise (Bad_fable_type "Attempt to open a labeled record");
		  match extract_row_opt omega value_type with(* type must be constrained already, since it requires an annotation *)
		      Some row -> 
			let (field_env, row_var), rec_row_var = unwrap_row row in
			let row_fields = Fable.row_labels_inorder field_env in(*  List.sort (fun (k1,_) (k2,_) -> compare k1 k2) (StringMap.to_alist field_env) in  *)
			let fields = 
                          List.fold_right 
                            (fun (k,s) a -> (* rearrange fields in binding order *)
			       let v = try List.assoc k fields
                               with Not_found -> Utility.gensym()
                               in
				 (k,v)::a) row_fields [] in
			let folder (omega, sigma, vtypes) (k1, var) (k2, spec) =
			  if (k1 <> k2) then 
			    raise (Bad_fable_type "Missing field in open record");
			  match spec with 
			      `Present t -> 
				begin
				  match Fable.get_name_bindings t.ftype with
				      [], None ->   (* duplicate name check already done in the pattern compiler *)
					let t' = Subst.apply sigma t in
					  Typeenv.add_normal_binding var ([],t') omega, sigma, (var, t')::vtypes
				    | [], Some n -> 
					let t' = Fable.strip_name_bindings_type t in
					let t' = Subst.apply sigma t' in
					let omega' = Typeenv.add_noshadow_binding var ([], t') omega in
					let sigma' = Subst.extend (n, Variable(var, `T(dummy_position, t', None))) sigma in
					  omega', sigma', (var, t')::vtypes
                                    | _ -> raise Impos
                                end
                            | `Absent -> raise (Bad_fable_type "Cannot open a record with absent fields") (* hmm ... maybe ok if k1 <> k2? *) in
                        let omega', _, ftypes = List.fold_left2 folder (omega, Subst.empty, []) fields row_fields in
                          type_check_expression omega' body, ftypes, fields
                    | _ -> raise (Bad_fable_type ("The open construct is reserved for dependent records; used here with " ^
                                                    (DerivingSyntax.Show_datatype.show value_type))) in
              let collapsed = Record_unpack(fields, value, body, None, `T(pos, type_of_expression body, None)) in
              let expansion = expand_record_open collapsed (List.rev ftypes) in (* desugar this for the interpreter *)
		check_no_escaping_names omega (type_of_expression body);
                Record_unpack(fields, value, body, Some expansion, `T(pos, type_of_expression body, None)) 

	  | Record_pack (v, value, dt, body, `U pos) -> 
	      let dt',_ = check_type omega dt in
	      let value' = type_check_expression (Typeenv.clear_type_annotation omega) value in
	      let _ = Unify.unify [] omega (Some value') false (type_of_expression value', dt') in
	      let omega' = Typeenv.add_normal_binding v ([], dt') omega in
	      let body' = type_check_expression omega' body in
		check_no_escaping_names omega (type_of_expression body');
		Record_pack(v, value', dt', body', `T(pos, type_of_expression body', None))
		  
          | Project (expr, label, `U pos) ->
              let expr = type_check_expression (Typeenv.clear_type_annotation omega) expr in
              let label_variable_type = annot_kindvar (fresh_type_variable ()).ltype MKind in
                unify (type_of_expression expr, annot_ukind (`Record (make_singleton_open_row (label, `Present (label_variable_type)))));
		if (Fable.free_names omega label_variable_type <> []) then 
		  raise (Bad_fable_type "Illegal projection from a dependent product; use open e1 as e2 instead");
                Project (expr, label, `T (pos, label_variable_type, None))
		  
          | Erase (value, label, `U pos) ->
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in (* Unify value type with an unlabeled record type *)
              let value_row = extract_row omega (type_of_expression value) in
                Erase (value, label, `T (pos, annot_ukind (`Record (row_with (label, `Absent) value_row)), None))

          | Variant_injection (label, value, `U pos) -> 
	      let omega = match Typeenv.get_type_annotation omega with 
		  None -> omega
		| Some t -> 
		    match (extract_row_opt omega t) with
			None -> raise (Bad_type_annotation (t, "row"))
		      | Some ((fsm, rv) as row)-> 
			  try 
			    match StringMap.find label fsm with 
				`Present t' -> Typeenv.set_type_annotation omega t'
			      | _ -> raise (Bad_type_annotation (t, "constructor " ^label))
			  with
			      Not_found -> 
				if (is_closed_row row) then 
				  raise (Bad_type_annotation (t, "constructor " ^label))
				else
				  Typeenv.clear_type_annotation omega in
              let value = type_check_expression omega value in
	      let vtype = type_of_expression value in
              let type' = `Variant (make_singleton_open_row (label, `Present vtype)) in
	      let maybe_label = match infer_label_type omega vtype with
		  `Not_label -> `Kind UKind
		| `General -> `Label(`Nothing, `Kind UKind)
		| `Specific -> 
		    let e = Variant_injection(label, value, `T(pos, annot_ukind type', None)) in
		    let pt = Unionfind.fresh (`Something e) in
		    let kv = Fable.fresh_kind_variable MKind in
		      `Label(`Inferred pt, `KindVar kv) in
		Variant_injection (label, value, `T (pos, Types.fable_annot type' maybe_label, None))
		  
          | Variant_selection (value, case_label, case_variable, case_body, variable, body, `U pos) ->
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in
              let value_type = type_of_expression value in
              let case_var_type = annot_kindvar (fresh_type_variable()).ltype MKind in
              let body_row = make_empty_open_row () in
              let variant_type = `Variant (row_with (case_label, `Present case_var_type) body_row) in
                unify (value_type, annot_kindvar variant_type UKind); 
                let case_body_omega = Typeenv.add_normal_binding case_variable ([], case_var_type) omega in
		let arg_expression = match case_body with 
		    Let("_", _, _, _) -> Syntax.unit_expression (`T (dummy_position, Types.unit_type(), None)) (* compensate for sugar.ml unit hack *)
		  | _ -> Variable(case_variable, `T(pos,case_var_type,None)) in
		let refinement = Variant_injection(case_label, arg_expression, 
						   `T(pos, value_type, None)) in
		let case_body_omega = Typeenv.add_refinement (value, refinement) case_body_omega in
                let case_body = type_check_expression case_body_omega case_body in
		  check_no_escaping_names omega (type_of_expression case_body);
                  (*
                    We take advantage of absence information to give a more refined type when
                    the variant does not match the label i.e. inside 'body'.

                    This allows us to type functions such as the following which fail to
                    typecheck in OCaml!

                    fun f(x) {
                    switch x {
                    case A(B) -> B;
                    case A(y) -> A(f(y));
                    }
                    }
                    
                    On the right-hand-side of the second case y is assigned the type:
                    [|B - | c|]
                    which unifies with the argument to f whose type is:
                    [|A:[|B:() | c|] |]
                    as opposed to:
                    [|B:() | c|]
                    which clearly doesn't!
                  *)
                  let body_var_type = annot_ukind (`Variant (row_with (case_label, `Absent) body_row)) in
                  let body_omega = Typeenv.add_normal_binding variable ([], body_var_type) omega in
                  let body = type_check_expression body_omega body in
                  let case_type = type_of_expression case_body in
                  let body_type = type_of_expression body in
		  let _ = unify (case_type, body_type) in
		    check_no_escaping_names omega body_type;
		    Variant_selection (value, case_label, case_variable, case_body, variable, body, `T (pos, body_type, None))
                      
          | Variant_selection_empty (value, `U pos) ->
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in
              let new_row_type = annot_ukind (`Variant (make_empty_closed_row())) in
                unify(new_row_type, type_of_expression value);
                Variant_selection_empty (value, `T (pos, fresh_type_variable (), None))
		  
          | Nil (`U pos) ->
	      let elt_t = annot_kindvar (fresh_type_variable()).ltype MKind in 
		Nil (`T (pos, annot_ukind (`Application ("List", [elt_t])), None))
		  
          | List_of (elem, `U pos) ->
	      let omega = match Typeenv.get_type_annotation omega with
		  None -> omega
		| Some t -> 
		    let rec get_elt_type t = match t.ltype with
		      | `Application("List", [elt_t]) -> elt_t
		      | `Application(alias, args) -> 
			  let vars, def = Typeenv.lookup_alias (alias, args) omega in
			  let t' = Instantiate.instantiate_alias (vars, def) args in
			    get_elt_type t'
		      | _ -> raise (Bad_type_annotation (t, "List")) in
		    let elt_t = get_elt_type t in
		      Typeenv.set_type_annotation omega elt_t
	      in
              let elem = type_check_expression (Typeenv.clear_type_annotation omega) elem in
                List_of (elem,
                         `T (pos, annot_ukind (`Application ("List", [type_of_expression elem])), None))

          | Concat (l, r, `U pos) ->
              let tvar = annot_kindvar (fresh_type_variable()).ltype MKind in 
              let l = type_check_expression omega l in
                unify (type_of_expression l, annot_ukind (`Application ("List", [tvar])));
                let r = type_check_expression omega r in
                  unify (type_of_expression r, type_of_expression l);
                  let type' = `Application ("List", [tvar]) in
                    Concat (l, r, `T (pos, annot_ukind type', None))

          | For (expr, var, value, `U pos) ->
              (* 	      print_string ("List comprehension: " ^(DerivingSyntax.Show_untyped_expression.show expression)^"\n"); *)
              let value_tvar = fresh_type_variable () in
              let expr_tvar = fresh_type_variable () in
	      let omega = match Typeenv.get_type_annotation omega with
		  None -> omega 
		| Some t -> match t.ltype with
		      `Application("List", [elt_t]) -> 
			Typeenv.set_type_annotation omega elt_t
		    | _ -> raise (Bad_type_annotation (t, "List")) in
              let value = type_check_expression (Typeenv.clear_type_annotation omega) value in
                unify (type_of_expression value, def_fable_annot (`Application ("List", [value_tvar])));
                let expr_omega = Typeenv.add_normal_binding var  ([], value_tvar) omega in
                let expr = type_check_expression (Typeenv.clear_type_annotation expr_omega) expr in
                  unify (type_of_expression expr, annot_ukind (`Application ("List", [expr_tvar])));
                  let type' = type_of_expression expr in
		    check_no_escaping_names omega type';
                    For (expr, var, value, `T (pos, type', None))
                      
          | Call_cc(arg, `U pos) -> 
              (* TBD: Make this a primitive function (need to pass c.c. to prims). *)
              let arg = type_check_expression omega arg in
              let contrettype = fresh_type_variable () in
              let anytype = fresh_type_variable () in
              let mailboxtype = instantiate omega "_MAILBOX_" in
              let conttype =
                annot_ukind (`Function (make_tuple_type [contrettype], mailboxtype, anytype)) in
              let argtype = annot_ukind (`Function (make_tuple_type [conttype], mailboxtype, contrettype)) in
                unify (argtype, type_of_expression arg);
                Call_cc(arg, `T (pos, contrettype, None))

          | Database (params, `U pos) ->
              let params = type_check_expression omega params in
                unify (type_of_expression params, db_descriptor_type);
                Database (params, `T (pos, annot_ukind (`Primitive `DB), None))

          | TableQuery (ths, query, `U pos) ->
              let row =
                (List.fold_right
                   (fun col env -> 
                      match col with 
                        | Left col -> StringMap.add col.Query.name (`Present col.Query.col_type) env
                        | Right _ -> env)
                   query.Query.result_cols StringMap.empty, Unionfind.fresh ClosedRow) in
              let datatype =  `Application ("List", [annot_ukind (`Record row)]) in
              let rrow = make_empty_open_row () in
              let wrow = make_empty_open_row () in
              let ths = alistmap (type_check_expression (Typeenv.clear_type_annotation omega)) ths
              in
                Utility.for_each ths 
                  (fun _, th -> 
                     ignore (unify (type_of_expression th, annot_ukind (`Table (annot_ukind (`Record rrow), annot_ukind (`Record wrow))))));
                unify_rows (row, rrow);
                TableQuery (ths, query, `T (pos, annot_ukind datatype, None))

          | TableHandle (db, tableName, tc, `U pos) ->
	      let (readtype,_), (writetype, _) = check_type omega tc.readtype, check_type omega tc.writetype in
	      let check_marshall_functions out (fieldname, e) = 
		(try 
		   let _ = List.assoc fieldname out in 
		     raise (Bad_fable_type ("Repeated field constraint for field " ^ fieldname))
		 with 
		     _ -> ());
		let rftype = match readtype.ltype with
		    `Record (rfsm, _) ->  
		      begin
			match (StringMap.lookup fieldname rfsm) with 
			    None
			  | Some `Absent -> raise (Bad_fable_type ("Marshalling constraint refers to a non-existent field : " ^fieldname))
			  | Some (`Present t) -> t
		      end
		  | _ -> raise (Unify_failure "Expected record types for a table row") in
		let mk_function_type arg ret =
		  let arg = Fable.strip_name_bindings_type arg in
		  let ret = Fable.strip_name_bindings_type ret in
		  let mb = fresh_type_variable () in
		    {ltype = `Function (make_tuple_type [arg], mb, ret); ftype = `Kind UKind} in
		let mtype = make_tuple_type [mk_function_type rftype (string_type());
					     mk_function_type (string_type()) rftype] in
		let e = type_check_expression omega e in 
 		let _ = unify (mtype, type_of_expression e) in
		  (fieldname, e)::out in
	      let marshallby = List.fold_left check_marshall_functions [] tc.marshallby in
              let datatype =  `Table (readtype , writetype) in
              let db = type_check_expression omega db in
              let tableName = type_check_expression omega tableName in
                unify (type_of_expression db, annot_ukind (`Primitive `DB));
                unify (type_of_expression tableName, string_type ()); 
                TableHandle (db, tableName, {readtype=readtype; writetype=writetype; marshallby=marshallby; primarykey=tc.primarykey}, `T (pos, annot_ukind datatype, None))

          | SortBy(expr, byExpr, `U pos) ->
              (* FIXME: the byExpr is typed freely as yet. It could have any
                 orderable type, of which there are at least several. How to
                 resolve this? Would kill for type classes. *)
              let byExpr = type_check_expression omega byExpr in
              let expr = type_check_expression omega expr in
                SortBy(expr, byExpr, `T (pos, type_of_expression expr, None))
                  
          | Wrong (`U pos) ->
              Wrong(`T (pos, fresh_type_variable(), None))

          | HasType(expr, datatype, `U pos) ->
              let datatype,_ = check_type omega datatype in
                free_alias_check omega datatype;
		let expr = type_check_expression (Typeenv.set_type_annotation omega datatype) expr in
		let expr_type = type_of_expression expr in
                  begin
                    try unify(expr_type, datatype);
                    with Unify_failure msg -> 
                      let _,_,src = position expr in
			type_mismatch
                          ~expected:datatype
                          ~inferred:expr_type
                          ~src:src
                          ~pos:pos msg
                  end;
                  HasType(expr, datatype, `T (pos, datatype, None))
      with
	| Bad_type_annotation(t,msg)-> raise (Type_error(position expression, "Incompatible type annotation: expected " ^msg^ "; got " ^(string_of_datatype t)))
	| FableType.Bad_fable_type msg
	| Typeenv.Shadow_variable msg
	| Bad_fable_type msg
        | Unify_failure msg
        | UndefinedVariable msg
        | UndefinedAlias msg ->
            raise (Type_error(position expression, msg))
  in (* end aux *)
    fun omega exp -> 
      let exp' = aux omega exp in
	match Typeenv.get_type_annotation omega with
	    None -> exp' 
	  | Some annot_typ -> 
	      let inferred_typ = Syntax.node_datatype exp' in
	      let _ = Unify.unify [] omega (Some exp') true (inferred_typ, annot_typ) in
		Syntax.set_node_datatype (exp', annot_typ) 
                  (* end "type_check" *)
		  
                  

(** type_check_mutually
    Companion to "type_check_expression"; does mutual type-inference

    [QUESTIONS]
    - what are the constraints on the definitions?
    - do the functions have to be recursive?
*)
and
    type_check_mutually omega (* (env, alias_env) *) (defns : (string * Syntax.untyped_expression * Types.u_datatype option) list) =
  let var_env = (map (fun (name, _, t) ->
                        match t with
                          | Some t ->
                              (* 			      print_string ("Ascribed u_type: " ^ (DerivingSyntax.Show_u_datatype.show t) ^ "\n"); *)
                              (* 			      flush stdout;	       *)
			      let t,_ = check_type omega t in 
                                (* 				print_string ("Ascribed type: " ^ (DerivingSyntax.Show_datatype.show t) ^ "\n"); *)
                                (* 				flush stdout; *)
			      let assumption = generalise omega t in
				(name, assumption)
                          | None -> (name, ([], fresh_type_variable ())))
                   defns) in
  let inner_omega = List.fold_left (fun omega (name, assumption) -> add_normal_binding name assumption omega) omega var_env in
  let type_check result (name, expr, u_topt) =
    let u_topt' = opt_app ((fun (x,_) -> Some x) -<- check_type omega) None u_topt in
    let inner_omega = Utility.opt_app (Typeenv.set_type_annotation inner_omega) inner_omega u_topt' in
    let expr = type_check_expression inner_omega expr in
    let t' = type_of_expression expr in
      match t'.ltype with
        | `Function _ as f  ->
            let t'' = snd (assoc name var_env) in
              Unify.unify [] inner_omega (Some expr) false (t', t'');
              (* 	      print_string ("Inferred type for abstraction " ^name^ ":" ^Types.string_of_datatype t'^"\n"); *)
              (* [HACK]
                 
                 This allows aliases to persist providing no
                 mailbox types have been instantiated.
              *)
              let expr = 
                if (Typeenv.is_mailbox_free inner_omega t') then
                  set_node_datatype (expr, t'')
                else
                  expr
              in
                (name, expr, u_topt') :: result
        | _ -> Errors.letrec_nonfunction (pos_of_expression expr) (expr, t') in

  let defns = fold_left type_check [] defns in
  let defns = rev defns in
    
  let omega' = List.fold_left
    (fun omega (name, value, _) -> 
       let assumption = 
         generalise omega (type_of_expression value) in
         Typeenv.add_normal_binding name assumption omega) omega defns in
    omega', defns     

let mutually_type_defs
    (env : Typeenv.omega)
    (defs : (string * Syntax.untyped_expression * 'a option) list)
    : (Typeenv.omega * (string * expression * 'c) list) =
  let new_env, new_defs = type_check_mutually env defs
  in (new_env, new_defs)
       

let type_expression : Typeenv.omega -> Syntax.untyped_expression -> (Typeenv.omega * expression) =
  fun typing_env exp ->
    typing_env, type_check_expression typing_env exp

let type_definition : Typeenv.omega -> untyped_definition -> (Typeenv.omega * definition) =
  fun omega (*env, alias_env*) def ->
    let omega', def' =
      match def with
        | Module (name, `U pos) -> 
            omega, Module (name, `T (pos, unit_type(), None))
        | Define (variable, value, loc, `U pos) ->
            let value = type_check_expression (Typeenv.set_context omega loc) value in
            let value_type = if is_value value then 
              (generalise omega (type_of_expression value))
            else [], type_of_expression value in
            let omega' = Typeenv.add_normal_binding ~location:loc variable value_type omega in
              omega', Define (variable, value, loc, `T (pos, type_of_expression value, None))
        | Alias (typename, vars, u_datatype, `U pos) ->
	    let datatype,_ = check_type omega u_datatype in
            let omega' = Typeenv.register_alias (typename, vars, datatype) omega in
              omega',
            Alias (typename, vars, datatype, `T (pos, annot_ukind (`Record (make_empty_closed_row ())), None))
        | Alien (language, name, (q, u_t),  `U pos)  ->
	    let t,_ = check_type omega u_t in
	    let assumption = (q, t) in
            let omega' = Typeenv.add_normal_binding name assumption omega in
              omega', Alien (language, name, assumption, `T (pos, snd assumption, None))
    in
      omega', def'
	
        
let type_program : Typeenv.omega -> untyped_program -> (Typeenv.omega * program) =
  fun typing_env (Program (defs, body)) ->
    let type_group (typing_env, typed_defs) : untyped_definition list ->
      Typeenv.omega * definition list = function
        | [x] -> (* A single node *)
            let typing_env, def = type_definition typing_env x in 
              typing_env, typed_defs @ [def]
        | xs  -> (* A group of potentially mutually-recursive definitions *)
            let defparts = map (fun (Define x) -> x) xs in
            let defbodies = map (fun (name, Rec ([(_, expr, t)], _, _), _, _) -> 
                                   name, expr, t) defparts in
            let (typing_env : Typeenv.omega), defs = mutually_type_defs typing_env defbodies in
            let defs = (map2 (fun (name, _, location, _) (_, expr, _) -> 
                                Define(name, expr, location, expression_data expr))
                          defparts defs) in
              typing_env, typed_defs @ defs

    and bothdefs l r = match l, r with
      | Define (_, Rec _, _, _), Define (_, Rec _, _, _) -> true
      | _ ->  false
    in
    let def_seqs = groupBy bothdefs defs in
    let mutrec_groups = (Callgraph.refine_def_groups def_seqs) in
    let typing_env, defs =
      fold_left type_group (typing_env, []) mutrec_groups in
    let typing_env, body = type_expression typing_env body in
      typing_env, (Program (defs, body))

(* Check for duplicate top-level definitions.  This probably shouldn't
   appear in the type inference module.

   (Duplicate top-level definitions are simply not allowed.)

   In future we should probably allow duplicate top-level definitions, but
   only if we implement the correct semantics!
*)
let check_for_duplicate_defs 
    omega
    (defs :  untyped_definition list) =
  let check (env, defined) = function
    | Define (name, _, _, `U position) when StringMap.mem name defined ->
        (env, StringMap.add name (position :: StringMap.find name defined) defined)
    | Define (name, _, _, `U position) when StringSet.mem name env ->
        (env, StringMap.add name [position] defined)
    | Define (name, _, _, _) ->
        (StringSet.add name env, defined)
    | _ -> 
        (env, defined) in 
  let env = Typeenv.all_bound_names omega in 
  let _, duplicates = List.fold_left check (env,StringMap.empty) defs in
    if not (StringMap.is_empty duplicates) then
      raise (Errors.MultiplyDefinedToplevelNames duplicates)

(* Given a datatype that contains no fable annotations other than 
   default typevariable kind annotations, this function walks over the type
   and massages the fable qualifiers into (Kind of tft * kind) qualifiers.
   This is used to initialize the type environment with the types of library functions *)
        

(* [HACKS] *)
(* two pass typing: yuck! *)
let type_program omega (Program (defs, _) as program) =
  check_for_duplicate_defs omega defs;
  Debug.if_set (show_typechecking) (fun () -> "Typechecking program...");
  type_program omega program

let type_expression omega expression =
  Debug.if_set (show_typechecking) (fun () -> "Typechecking expression...");
  type_expression omega expression
