(** Optimisation passes: rewrites of the AST. *)

open List

open Utility
open Rewrite
open Syntax
open Expression
open Expression.Expr
open Sql_transform

let optimising = Settings.add_bool("optimising", true, `User)
let show_opt_verbose = Settings.add_bool("show_opt_verbose", false, `User)
let show_optimisation = Settings.add_bool("show_optimisation", false, `User)
let reduce_recs = Settings.add_bool("reduce_recursion", false, `User)

(**
   Check that "Project" and "Erase" do not occur in the syntax tree.
*)
let no_project_erase : RewriteSyntax.rewriter = function
  | Project _ 
  | Erase _ -> assert false
  | _ -> None

(** [pure]

    Checkes whether the evaluation of an expression is known to be free
    from side effects
*)
let pure : expression -> bool = 
  (* Everything is pure except the application of certain primitive
     functions and of any functions which call those.  For now, we'll
     just punt when we see a function application.  Eventually the
     type system will help us out.

     NB: continuation invocation is impure in the sense that we can't
     replace `x = f(3); 4' with `4' if `f' is a continuation.
  *)
  let rec pure default = function 
      (* TBD: annotate ALL prim funcs as to pureness *)
    | Apply((Variable("take", _) | Variable("drop", _)), arg, _)
      -> for_all (pure default) arg
    | Apply _    -> false
    | App _       -> false
    | TableQuery _ -> false
        (* Is callCC pure? *)
    | e       -> default e
  and all_true l = fold_right (&&) l true in
    reduce_expression pure (all_true -<- snd)


(** Inlining **)

(* Number of nodes in a syntax tree *)
let countNodes e = 
  let count = ref 0 in 
    Syntax.reduce_expression
      (fun recurse e -> incr count; recurse e)
      (fun _ -> ()) e; 
    !count;;

(* Inline small, non-recursive functions *)
let contains_no_extrefs : Syntax.expression -> bool =
  (=) [] -<- List.filter (not -<- flip List.mem_assoc (fst (Typeenv.extract_typing_environment Library.type_env))) -<- StringSet.elements -<- freevars

let recursivep : Syntax.expression -> bool = function
  | Rec ([(name, fn, _)], Variable (v, _), _) when v = name 
      -> StringSet.mem name (freevars fn)
  | _ -> false

let size_limit = 150

let is_inline_candidate' = function
  | Define (_, (Rec _ as e), _, _) -> 
      not (recursivep e) && contains_no_extrefs e && countNodes e < size_limit
  | Define (_, e, _, _) when Syntax.is_value e -> contains_no_extrefs e && pure e
  | _ -> false

let find_inline_candidates es : (string * expression * location) list = 
  let is_inline_candidate = function
    | Define (name, rhs, location, _) as e when is_inline_candidate' e -> 
        [name, rhs, location]
    | _ -> []
  in Utility.concat_map is_inline_candidate es

let location_matches location = function
  | Define (_, _, location', _) -> location=location'
  | _ -> false

let perform_value_inlining location name rhs (Program (defs, body)) =
  Program (
    List.map (fun def ->
                if location_matches location def then
                  Syntax.subst_fast_def name rhs def
                else
                  def) defs,
    Syntax.subst_fast name rhs body)

let perform_function_inlining location name var rhs (Program (defs, body)) =
  let replace_application : RewriteSyntax.rewriter = function
    | Apply (Variable (n, _), [p], d) when n = name -> 
        (* FIXME: inlining only implemented for single-argument functions! *)
        Some (Let (var, p, rhs, d))
    | _ -> None
  in
    Program (
      List.map
        (fun def ->
           if location_matches location def then
             Syntax.rewrite_def (RewriteSyntax.bottomup replace_application) def
           else
             def) defs,
      fromOption body (RewriteSyntax.bottomup replace_application body))

let inline (Program (defs, body) as program) = 
  let valuedefp = function
    | _, Rec _, _ -> false
    | _        -> true
  in
  let candidates = find_inline_candidates defs in
  let value_candidates, fn_candidates = List.partition valuedefp candidates in
  let program' = 
    List.fold_right 
      (fun (name, rhs, location) program ->
	 perform_value_inlining location name rhs program)
      value_candidates
      program
  in 
  let program'' = 
    List.fold_left 
      (fun program (_, rhs, location)  ->
         match rhs with
           | Rec ([(name, Abstr ([Variable(v, _)], body, _), _)], _, _) ->
	       perform_function_inlining location name v body program
           | Rec _ ->
                program)
      program'
      fn_candidates
  in program''

(* Minimize the amount of recursion expressed in the syntax tree *)
let reduce_recursion : RewriteSyntax.rewriter = function
  | Rec (bindings, cont, data) ->
      let untyped_bindings = List.map (fun (name, expr, _) -> name, expr) bindings in
      let find_definition v = List.find (fun (name, expr, annotation) -> name = v) bindings in
      let recursive_p (name, expr, annot) =  StringSet.mem name (freevars expr) in
      let cliques = Callgraph.group_and_order_bindings_by_callgraph untyped_bindings in
        begin match map (map find_definition) cliques with
          | [_::_::_] -> None (* one multi-element group.  Everything must be mutually-recursive *)
          | [[x]] when recursive_p x -> None (* One single-element group; element is recursive. *)
          | groups -> (* either a group consisting of a single-element non-recursive element 
                         or multiple groups *)
              Some (List.fold_right
                      (fun group body -> 
                         match group with 
                           | [(name, expr, annot)] when not (StringSet.mem name (freevars expr)) ->
                               let rhs = match annot with
                                 | Some annot -> HasType (expr, annot, data)
                                 | None -> expr in
                                 Let (name, rhs, body, data)
                           | xs -> Rec (xs, body, data)) groups cont) end
  | _ -> None



(** [uniquify_names]

    Give unique names to all local bindings.  Local names should be
    distinct from each other and from toplevel names

    After this, we can be much less careful about scope.
*)
let uniquify_names : RewriteSyntax.rewriter = 
  let rewrite_node = function
    | Abstr (vs, b, data) -> 
        let names = List.map (fun (Variable(v, data)) -> (v, Variable(gensym ~prefix:v (), data))) vs in
          Some (Abstr (List.map snd names, 
                      List.fold_right 
                        (uncurry Syntax.rename_fast)
                         (List.map (fun (v,Variable(v',_)) -> (v,v')) names)
                         b, data))
    | Let (v, e, b, data) -> 
        let name = gensym ~prefix:v () in
          Some (Let (name, e, Syntax.rename_fast v name b, data))
    | Rec (vs, b, data) -> 
        let bindings = List.map (fun (name, _, _) -> 
                                   (name, gensym ~prefix:name ())) vs in
        let rename = List.fold_right (fun (x, r) expr ->
                                        Syntax.rename_fast x r expr) bindings in
          Some(Rec(List.map (fun (n, v, t) ->
                               (List.assoc n bindings, rename v, t)) vs,
                   rename b, data))
    | Record_selection (lab, lvar, var, value, body, data) ->
        let lvar' = gensym ~prefix:lvar ()
        and var'  = gensym ~prefix:var () in
          Some(Record_selection(lab, lvar', var', value, 
                                Syntax.rename_fast var var' (Syntax.rename_fast lvar lvar' body),
                                data))
    | For (b, v, src, data) -> 
        let name = gensym ~prefix:v () in
          Some (For(Syntax.rename_fast v name b, name, src, data))
    | Variant_selection (value, clab, cvar, cbody, var, body, data) ->
        let cvar' = gensym ~prefix:cvar ()
        and var'  = gensym ~prefix:var () in
          Some (Variant_selection (value, clab, 
                                   cvar', Syntax.rename_fast cvar cvar' cbody, 
                                   var',  Syntax.rename_fast var var' body,
                                   data))
    | _ -> None
  (* Note that this will only work bottomup, not topdown, since
     we need to replace bindings from the inside out *)
  in RewriteSyntax.bottomup rewrite_node

(*
(** [inference_rw]
    The type-inference function, in the form of a rewriter
*)
let inference_rw env : RewriteSyntax.rewriter = fun input -> 
  let output = snd (Inference.type_expression env (erase input)) in
    if input = output then None
    else Some output
*)

(** {0 Renaming} *)

(** renaming
    I think this is meant to "remove renamings", that is places where the user
    has written { var x = y; ... }
    It is incomplete, becuase the guard is very conservative. If we used
    capture-avoiding substitution, or ensured binder uniqueness in advance,
    we would be more complete.
*)
let renaming : RewriteSyntax.rewriter = 
  let bound_in var = 
    (* Is a particular name bound inside an expression? *)
    let binds default = function
      | Let (v, _, _, _)
          when v = var -> true
      | Abstr (vs, _, _) when List.exists (fun (Variable(v, _)) -> var = v)  vs -> true
      | Record_selection (_, v1, v2, _, _, _)
      | Variant_selection (_, _, v1, _, v2, _, _) when var = v1 || var = v2 -> true
      | Rec (bindings, _, _) when List.exists (fun (v,_,_) -> v = var) bindings -> true
      | other -> default other
    and combiner l = fold_left (||) false l 
    in reduce_expression binds (combiner -<- snd) 
  in function 
        (* 1. Don't replace where there's an inner binding of the name `x'.
           2. Don't replace where there's an inner binding of the name `y'.
        *)
  | Let (y, Variable (x,  _), body,  _) when bound_in x body || bound_in y body
      -> None (* Could do better: does the binding shadow? *)
  | Let (y, Variable (x, _), body, _)
      -> Some (Syntax.rename_fast y x body)
  | _ -> None

(** [unused_variables]: Remove all let bindings where the name is not used 
    in the body.

    FIXME: also letrec ("transitively") and record selection
    operators.

    FIXME: needs to take account of variables used within Query
    expressions.
*)

let unused_variables : RewriteSyntax.rewriter = function
    (* This'd be quite a bit more efficient bottom-up, passing the
       list of free variables outwards rather than searching the body
       every time we find a let. *)
  | Let (var, expr, body, _) when pure expr 
                               && not (StringSet.mem var (freevars body)) -> Some body
  | _ -> None
      (* FIXME: this ignores variables that are hidden inside queries *)



(** [simplify_regex] rewrites an expression as follows:
{
   s ~ let x in b
       let x in s ~ b
}

    (FIXME; Note there are lots of constructors that commute with let,
    and it might be quite wise to do so).

    In a regex match such as the following, the order of eval is unspecified:
    f() ~ /{g()}/
*)

let simplify_regex : RewriteSyntax.rewriter = function
  | Apply (Variable ("~", _) as tilde, [lhs; Let (v, e, rhs, d1)], d2)  ->
      Some (Let (v, e, Apply (tilde, [lhs;rhs], d1), d2))
  | _ -> None

(** {3 SQL utility values} These values are provided to ease the writing of SQL optimisers. *)

(** extract_tests
    Finds tests that can be pushed to a SQL query. Such tests are of the form '{i if cond then collection else [col] }'.
    param forbidden A list of variables that should not be used in the tests.
    @param bindings Bindings from variables to database fields.
    @param expr The expression from which the test must be extracted.
    @return A tuple (positive, negative, result, origin).
        `positive' and `negative' are query expressions. `result' is the 
        input expression with these tests removed. `origin` is a list of 
        record selections that need to be applied.
 *)
(*TODO: Is there any reason why positive and negative should be kept separate? *)
(*TODO: The calculated expression for the collection extension case
   is not valid *)
let rec extract_tests (bindings:bindings) (expr:expression)
    : (Query.expression list * expression * projection_source list) =
  match expr with
    | Let (_, Variable (_, _), _, _) ->
        failwith "Internal error: renaming declarations should have been removed by earlier optimisations"
    | Let (variable, value, body, data) ->
        let (condns, body, origin) = extract_tests (`Unavailable variable :: bindings) body in
          (condns, Let (variable, value, body, data), origin)

    | Record_selection (label, label_variable, variable, Expression.Expr.Variable (name, vdata), body, data) ->
        let trace_data = `Selected{field_name=label; field_var=label_variable;
                                   etc_var = variable; source_var = name} in
        let (condns, body, origin) = extract_tests (trace_data :: bindings) body in
          (condns,
	   Record_selection (label, label_variable, variable, 
			     Expression.Expr.Variable (name, vdata), body, data), 
	   origin)
    | Record_selection (label, label_variable, variable, value, body, data) ->
        let condns, body, origin = extract_tests (`Unavailable variable :: bindings) body in
          (condns, Record_selection (label, label_variable, variable, value, body, data), origin)

    | Condition (condition, t, ((Nil _) as e), data)  ->
        let condns, t, origin = extract_tests bindings t in
          (match condition_to_sql condition bindings with
             | Some (sql_condition, new_origin) ->
                 (sql_condition:: condns, t, new_origin @ origin)
             | None ->
                 (condns, Condition(condition, t, e, data), origin))
    | Condition (condition, ((Nil _) as t), e, data) ->
        let condns, e, origin = extract_tests bindings e in
          (match condition_to_sql condition bindings with
             | Some (sql_condition, new_origin) ->
                 (Sql.negation sql_condition:: condns, e, new_origin @ origin)
             | None ->
                 (condns, Condition(condition, t, e, data), origin))
    | For (expr, variable, value, data) ->
        let condns, expr, origin = 
          extract_tests (`Calculated (variable, expr) :: bindings) expr in
          (condns, For (expr, variable, value, data), origin)

    | _ -> ([], expr, [])

exception Lift_app_failure 
let rec lift_applications (bindings:bindings) (expr:expression) 
    : (query * Syntax.expression) option =
  let module E = Expression.Expr in
  let module Q = Query in 
  let loopvar, query = match List.find (function `Table_loop(_, _) -> true | _ -> false) bindings with
      `Table_loop(loopvar, query) -> loopvar, query
    | _ -> raise Utility.Impos in
  let rec aux bindings expr =  match expr with 
      
    | E.Record_selection (label, label_variable, variable, E.Variable (name, vdata), body, data) ->
        let trace_data = `Selected{field_name=label; field_var=label_variable;
                                   etc_var = variable; source_var = name} in
	let qb_opt = aux (trace_data :: bindings) body in
	  (match qb_opt with 
	       None -> None
	     | Some (q,body') -> Some (q, Record_selection (label, label_variable, variable, E.Variable (name, vdata), body', data)))
		
    | E.Record_selection (label, label_variable, variable, value, body, data) ->
        let qb_opt = aux (`Unavailable variable :: bindings) body in
	  (match qb_opt with 
	       None -> None
	     | Some (q,body') -> Some (q, Record_selection (label, label_variable, variable, value, body', data)))

    | E.Let(letvar, (E.Apply(E.Variable(fname, `T(_, _, Some "policy")), fargs, _) as app), body, ldata) -> 
	
	let liftable = function
	    E.Variable(name, _) -> match trace_variable name bindings with
		`Table_field (tabalias, colname) -> 
		  let tabinst = query.Q.tables in 
		    List.exists (fun (_, alias) -> tabalias = alias) tabinst
	      | `Earlier (_, []) -> true
	      | _ -> false  in 
	  if not (List.for_all liftable fargs) then None
	  else 
	    begin
 	      match make_sql bindings app with
		  Some(appsql, _) ->
		    let column_alias = Utility.gensym ~prefix:"Exprcol_" () in
		    let table_alias = match query.Q.tables with 
			[(_, alias)] -> alias
		      | _ -> raise  Lift_app_failure in
		    let exprcol = {Q.expression = appsql; Q.exprcol_alias = column_alias; Q.exprcol_table_alias = table_alias} in
		    let query = {query with Q.result_cols = query.Q.result_cols @ [inRight exprcol]} in
		    let bindings = List.fold_right (fun b bl -> match b with
							`Table_loop _ -> `Table_loop(loopvar, query)::bl
						      | _ -> b::bl) bindings [] in
		    let dummy_var = Utility.gensym ~prefix:"ignore_" () in
		    let dummy_typed_data =  `T (dummy_position, Types.unit_type(), None) in
		    let expr = E.Record_selection(column_alias, letvar, dummy_var, E.Variable(loopvar, dummy_typed_data), body, ldata) in
		      lift_applications bindings expr
		| None -> None
	    end
	      
    | _ -> Some(query, expr) 
  in
    try 
      match expr with 
	  E.Condition(where, tt, ff, data) -> 
	    (match aux bindings where with 
		 None -> None 
	       | Some (q, where') -> Some(q, E.Condition(where', tt, ff, data)))
	| _ -> aux bindings expr
    with
	Lift_app_failure -> None
		
	  
(** {3 SQL optimisers} All following values are optimiser functions
    that can be applied to an expression. Generally, they try to push
    as much calculation to the DBMS as possible. {e Beware:} some
    optimisers have dependencies with other optimisers. *)

(** For every collection extension that has a table as source,
    modifies the SQL query associated to that table to only select
    fields that will be used in the body of the extension. {e Beware:}
    this version of the algorithm considers a field as used if a
    record selection operators extracts it from the record
    representing the relation from the database, even if it is not
    used in the selection's body. *)
type fieldset = All
		| Fields of (string list)

(* FIXME: Make sure we don't throw away fields that are used in 
   a SortBy clause. *)
let sql_projections : RewriteSyntax.rewriter =
  
  let merge_needed : fieldset list -> fieldset =
    let merge2 = function
      | All, _ | _, All -> All
      | Fields x, Fields y -> Fields (x @ y) in
      fold_left (curry merge2) (Fields []) in

  let needed_fields (var:string) : expression -> fieldset =

    let rec visitor (vars:string list) default : expression -> fieldset = function
        | Variable (name, _) when List.mem name vars -> All
	    
        | Record_selection (label, _, variable, Variable (name, _), body, _) when List.mem name vars ->
            merge_needed (Fields [label] :: (visitor vars default body) :: [visitor (variable::vars) default body])
	      
        | other -> default other
	    
    in reduce_expression (visitor [var]) (merge_needed -<- snd) in

  let rewrite = function
    | For (body, variable, TableQuery(th, query, tdata), data) ->
	let nf = needed_fields variable body in
          (match nf with
             | Fields needs ->
		 Some (For (body, variable,
                            TableQuery(th, 
                                       project needs query,
                                       tdata),
                            data))
             | All -> None)
    | _ -> None in
    
    RewriteSyntax.bottomup rewrite

(** For every collection extension that has a table as source,
    modifies the SQL query associated to that table to add as many
    conditions from the extension's body to the query itself as
    possible. To be considered for optimisation, a condition must be
    of the form '{i if cond then [] else ...}' or '{i if cond then
    ... else []}', where '{i cond}' is a comparison between a
    constant and a variable or between two variables (coming from the
    database). The optimisation will stop looking for conditions in
    the body expression if it encouters any other node than a
    condition, a {i let}, a record selection or a collection
    extension. *)
let sql_selections : RewriteSyntax.rewriter = function 
  | For (expr, variable, TableQuery(th, query, tdata), data) ->
(*       print_string ("sql_selections: query \n " ^ Query.Show_query.show query ^ "\n"); *)
(*       print_string ("sql_selections: where clause \n " ^ Expression.DerivingSyntax.Show_debug_expression.show expr ^ "\n"); *)
(*       flush stdout; *)
      let query, expr = 
	if (Settings.get_value Sql_transform.call_stored_procs) then
	  match lift_applications [`Table_loop (variable, query)] expr with
	      None -> query, expr
	    | Some (q,e) -> q,e
	else query, expr in
(* 	print_string "After lifting \n"; *)
(* 	print_string ("sql_selections: query \n " ^ Query.Show_query.show query ^ "\n"); *)
(* 	print_string ("sql_selections: where clause \n " ^ Expression.DerivingSyntax.Show_debug_expression.show expr ^ "\n"); *)
(* 	flush stdout; *)
	let condns, expr, origin = extract_tests [`Table_loop (variable, query)] expr in
	let table = TableQuery(th, select condns query, tdata) in
          Some (select_by_origin origin (For (expr, variable, table, data)))
  | _ -> None


let substitute_projections' new_src renamings bindings expr =
  let subst_projection (from, to') : RewriteSyntax.rewriter = function
    | Record_selection (label, label_var, etc_var, Variable (src, d), body, 
                        data) as orig
        when from.Query.col_alias = label ->
        Debug.if_set show_optimisation
          (fun () -> "Renaming " ^ from.Query.col_alias ^ " to " ^ to' ^ " in " ^ string_of_expression orig);
        (match trace_variable src bindings with
	   | `Table query when mem from.Query.table_alias (map snd query.Query.tables) ->
               Some(Record_selection(to', label_var, etc_var, 
                                     Variable (new_src, d), body, data))
           | `Table_field (table_as, _) when from.Query.table_alias = table_as ->
               (* NOTE: I think this case never occurs *)
               Some (Record_selection (to', label_var, etc_var, Variable (new_src, d), body, data))
           | `Unavailable -> Debug.if_set show_optimisation
               (fun () -> src ^ " was unavailable."); None
           | `Earlier(str, _) -> Debug.if_set show_optimisation (fun () -> "Earlier: " ^ str); None
           | _ -> 
               Debug.if_set show_optimisation (fun () -> "NOT renaming after all!");
               Some (Record_selection (label, label_var, etc_var, 
                                       Variable (src, d), body,  (* not new_src? *)
                                       data)))
    | _ -> None
  in
    RewriteSyntax.all (List.map (fun r -> RewriteSyntax.bottomup (subst_projection r)) renamings) expr

let substitute_projections new_src renamings bindings expr = 
 do_rewrite (substitute_projections' new_src renamings bindings) expr

let read_proj = function
    Record_selection(field, _, _, record, 
                     Variable _, _) ->
      Some(record, field)
  | _ -> None

let rec sql_sort = function
  | SortBy(TableQuery(th, query, data1), 
           Abstr([Variable(loopVar,_)], sortByExpr, _), _) ->
      (match read_proj sortByExpr with
           Some (Variable(sortByRecVar, _), sortByFld)
             when sortByRecVar = loopVar
               -> Some(TableQuery(th, Query.add_sorting query 
                               (`Asc(Query.owning_table sortByFld query,
				     sortByFld)), data1))
         | _ -> None)
  | _ -> None

let sql_aslist : RewriteSyntax.rewriter =
  function 
    | Apply(Variable("asList", _), [th], (`T (pos,_,_) as data)) ->
        let th_type = Types.concrete_type (node_datatype th) in
        let th_row = match th_type.Types.ltype  with
          |  `Table (rec_type,  _) -> begin
            match rec_type.Types.ltype with
			    `Record th_row -> th_row
              | `MetaTypeVar point -> begin
                match Unionfind.find point with
                    (* hack to get labeled TableHandles to work *)
                  | Types.Body ({Types.ftype=_;Types.ltype=`Record th_row}) 
                    -> th_row
                  | _ -> 
                    failwith "Internal Error: Unhandled table record."
                end
		      | _ -> failwith "Internal Error: rows must have record type"
		     end
          | _ -> failwith "Internal Error: tables must have table type"
        in
        let table_alias = gensym ~prefix:"Table_" () in
	let rowFieldToTableCol colName = function
	  | `Present fieldType -> {Query.table_alias = table_alias; 
                                   Query.name = colName; 
				   Query.col_alias = colName; 
                                   Query.col_type = fieldType}
	  | _ -> failwith "Internal Error: missing field in row"
	in
	let fields, _ = th_row in
	let columns = StringMap.to_list rowFieldToTableCol fields in
        let th_var = match th with
          | Variable(var, _) -> var
          | _ -> gensym ~prefix:"_t" () in
	let select_all = {Query.distinct_only = false;
			  Query.result_cols = List.map inLeft columns;
			  Query.tables = [(`TableVariable th_var,
                                           table_alias)];
			  Query.condition = Query.Boolean true;
			  Query.sortings = [];
			  Query.max_rows = None;
			  Query.offset = Query.Integer (Num.Int 0);
			  Query.projections = [];} in
        let th_list_type = `Application ("List", [Fable.annot FableType.UKind (`Record(th_row))]) in
        let table_query = TableQuery ([table_alias, 
                                       Variable (th_var, `T (pos, th_type, None))],
                                      select_all,
                                      `T (pos, Fable.annot FableType.UKind th_list_type, None))
        in
          (match th with
             | Variable _ -> Some table_query
             | _ -> Some (Let (th_var, th, table_query, data)))
    | _ -> None

(** check_join
    Inspects an expression for possible collection extension
    operators that can be joined with some outer comprhsn, the
    variable of the outer comprhsn indicated by the `loop_var' param.

    @param forbid Variables that if used in the body of an inner 
    extension prevent the optimisation
    @param db The name of the outter database. Only tables of the 
           same database can be joined.
    @param bindings Bindings of variables to rows or fields from a table.
    @param expr The expression to inspect.
    @return Either None or Some of a tuple containing: 
      * positive join conditions (conditions between fields of the inner 
        and outer table);
      * negative join conditions;
      * the inner query to join;
      * the new Links-AST expression without the joinable collection extensions. 
*)
let rec check_join (loop_var:string) (bindings:bindings) (expr:expression)
    =
  let bindings, expr = sep_assgmts bindings expr in
    match expr with
      | Condition (condition, t, ((Nil _) as e), data)  ->
          (match check_join loop_var bindings t with
	     | Some (condns, th, query, projs, var, t) ->
                 Some (condns, th, query, projs, var, 
                       Condition (condition, t, e, data))
	     | None -> None)
      | Condition (condition, ((Nil _) as t), e, data) ->
          (match check_join loop_var bindings e with
	     | Some (condns, th, query, projs, var, e) ->
                 Some (condns, th, query, projs, var,
                       Condition (condition, t, e, data))
	     | None -> None)
      | For (expr, variable, TableQuery(th, query, _), _) ->
          (* TODO: Test whether both tables come from the same database. *)
          (match extract_tests (`Table_loop (variable, query) :: bindings) expr with
               (*  ([], _, _) -> None *)
	     | (condns, expr, origin) ->
                   Some (condns, th, query, 
                         origin, variable, expr))
      | _ -> None

(** sql_joins
    When a collection extension has a table as source, explore the
    body expression for other collection extensions on tables from the
    same database. If a suitable extension is found, the query of the
    outer extension is modified to query on a join between the union
    of their tables, and the inner extension is removed.
*)
let rec sql_joins : RewriteSyntax.rewriter = 
  function
    | For (body, outer_var, (TableQuery (outer_ths, query, tdata)), data) ->
        let bindings = [`Table_loop (outer_var, query)] in
        (match check_join outer_var bindings body with
           | Some(condns, inner_ths, inner_query, origins, inner_var, body) ->
               let renamings, query = join condns (query, inner_query) in
                 
               (* Replace anything of the form inner_var.field with 
                  outer_var.renamed_field with renamings as given by 
                  the join operator *)
               let body = (substitute_projections outer_var renamings 
                             [`Table_loop(inner_var, inner_query)] body) in
               let expr = For(body,
                              outer_var, 
                              TableQuery (inner_ths @ outer_ths, query, tdata), 
                              data) in
(*                (\* finally, wrap the whole expression in the  *)
(*                   projections returned from check_join; *)
(*                   TBD: Does this need to go somewhere? Inside the loop? *\) *)
(*                let expr = select_by_origin origins expr in *)
                 
                 Some expr

           | None -> None

(*            | None ->  *)
(*                (\* check_join returned None, so perhaps we only have one loop.  *)
(*                   still, try to push the conditions down into SQL.  *)
(*                   HACK ALERT; this shouldn't be a special case. the 1-join *)
(*                   should be handled the same as 2-join, 3-join, etc. *\) *)
(*                let (pos, neg, body, proj_srcs) = extract_tests bindings body in *)
(*                  (\* This positive/negatives business is retarded, I think *\) *)
(*                  if (pos <> [] || neg <> []) then *)
(*                    let query = {query with Query.condition = pos_and_neg (query.Query.condition::pos, neg) } in *)
(*                      Some (For(body, outer_var,  *)
(*                                TableQuery(th, query, tdata), data)) *)
(*                  else None *)
        ) (* match check_join .... with *)
    | _ -> None

let lift_lets : RewriteSyntax.rewriter = function
  | For(loopbody, loopvar, Let(letvar, letval, letbody, letdata), data) 
    -> Some(Let(letvar, letval,
                For(loopbody, loopvar, letbody, data), letdata))
  | For(Let(letvar, letval, letbody, letdata), loopvar, src, data)
      when not (StringSet.mem loopvar (freevars letval))
        && pure letval
        -> Some(Let(letvar, letval, For(letbody, loopvar, src, data), letdata))
  | Condition(cond, Let(letvar, letval, letbody, letdata), e, data)
      when pure letval
        -> Some(Let(letvar, letval, Condition(cond, letbody, e, data), letdata))
  | Condition(cond, t, Let(letvar, letval, letbody, letdata), data)
      when pure letval
        -> Some(Let(letvar, letval, Condition(cond, t, letbody, data), letdata))
  | Apply(Let(letvar, letval, letbody, letdata), ap_arg, data)
        -> Some(Let(letvar, letval, Apply(letbody, ap_arg, data), letdata))
  | Apply(ap_func, [Let(letvar, letval, letbody, letdata)], data)
      when pure ap_func
        -> Some(Let(letvar, letval, Apply(ap_func, [letbody], data), letdata))
  | SortBy(Let(letvar, letval, letbody, letdata), byExpr, data) ->
      Some (Let(letvar, letval, SortBy(letbody, byExpr, data), letdata))
  | _ -> None


(* return true if the argument is an atom *)
let is_atom = function
  | Variable _ | Constant(Integer _, _) -> true
  | _ -> false

(* if e is not an atom then bind it to a variable by extending the
   continuation k return the new continuation and an atomic expression
   representing e
 *)
let lift_let data e k =
  if is_atom e then
    (fun body -> k body), e
  else
    let x = gensym () in
      (fun body -> k (Let (x, e, body, data))), Variable (x, expression_data e)



(** (1 take/drop optimization).
    Push calls to take and drop that surround queries into the query.
    [N.B. these rewrite rules play fast and loose with the `data'
    component of expression nodes.  Don't assume anything about the
    data after these have run.]
*)

(** [simplify_takedrop]
    The rewritings are as follows:
       take e1 (drop e2 e3) ~>  {x = e2; y = e1; take y (drop x e3)}
    (Not performed if both e1 and e2 are variables or integer literals)
       take e1 e2 ~> {x = e1; take x e2}
       drop e1 e2 ~> {x = e1; drop x e2}
    (Not performed if e1 is a variable or integer literal)
*)
let simplify_takedrop : RewriteSyntax.rewriter = function
  | Apply (Variable ("take"|"drop" as f, d1), [e1; e2]
             (*Record_intro (fields1, None, d2)*), d3) ->
      let k, e1 = lift_let d3 e1 identity in
        begin
          match f, e2 with
            | "take", Apply (Variable ("drop", d4), [e2; e3] (* Record_intro(fields2, None, d5) *), d6) ->
                let k, e2 = lift_let d3 e2 k in
                  Some(k (Apply (Variable ("take", d1),
                                 [e1; Apply (Variable ("drop", d4), [e2; e3], d6)], d3)))
            | _ -> Some (k (Apply (Variable (f, d1), [e1; e2], d3)))
        end
  | _ -> None

(** [push_takedrop] actually pushes [take] and [drop] calls into a query.
    Rewrites as follows: {[
        take e1 (drop e2 (Table (... q ...))) ~> Table (... {q with offset = e2; limit = e1} ...)
    }] where e1 and e2 are variables or integer literals
    {[
        take e1 (Table (... q ...)) ~> Table (... {q with limit  = e1} ...)
        drop e1 (Table (... q ...)) ~> Table (... {q with offset = e1} ...)
    }] where e1 is a variable or integer literal
*)
let push_takedrop : RewriteSyntax.rewriter = 
  let queryize = function
    | Variable (v, _) -> Query.Variable v
    | Constant(Integer  n, _) -> Query.Integer n
    | _ -> failwith "Internal error during take optimization" in 
  function
    | Apply (Variable ("take", _) as takef, [take_args1; take_args2], calltaked) ->
        if is_atom take_args1 then
          begin
            match take_args2 with
              | TableQuery (e2, q, d) ->
	          Some (TableQuery (e2, {q with Query.max_rows = Some (queryize take_args1)}, d))
              | Apply (Variable ("drop", _) as dropf, [drop_args1; drop_args2], calldropd) ->
                  if is_atom drop_args1 then
                    begin
                      match drop_args2 with
                        | TableQuery(ths, q, d) ->
                            Some(TableQuery(ths,
                                            {q with
                                               Query.max_rows = Some (queryize take_args1);
                                               Query.offset   = queryize drop_args1}, d))
                        | _ -> None
                    end
                  else None
              | For(List_of(expr, ldata) as body, var, src, ford) 
                  when pure(expr) ->
                  Some (For(body, var,
                            Apply (takef,
                                   [take_args1; src],
                                   calltaked), 
                            ford))
              | _ -> None
          end
        else None
    | Apply (Variable ("drop", _) as dropf, [drop_arg1; drop_arg2], d2) ->
        if is_atom drop_arg1 then
          begin
            match drop_arg2 with
              | TableQuery (ths, q, d) ->
  	          Some (TableQuery(ths, {q with Query.offset = queryize drop_arg1}, d))
              | For(List_of(expr, ldata) as body, var, src, fordata) 
                  when pure(expr) ->
                  Some (For(body, var,
                            Apply (dropf,
                                   [drop_arg1; src], 
                                   d2), 
                            fordata))
              | _ -> None
          end
        else None
    | _ -> None

let remove_trivial_extensions : RewriteSyntax.rewriter = function
  | For (List_of (Variable (v1, _), _), v2, e, _)
      when v1 = v2 -> Some e
  | _ -> None

let remove_unpacks : RewriteSyntax.rewriter = function
  | Record_unpack(_,_,_, Some e, _) -> 
      Some e 
  | _ -> None
     
(* Evaluate expressions involving only constants and pure functions at
 * compile time. *)
let fold_constant : RewriteSyntax.rewriter = 
  (* TODO: Also arithmetic, etc. *)
  let constantp = function
    | Constant _ | Nil _ -> true
    | Record_intro (fields, None, _) when StringMap.is_empty fields -> true
    | _ -> false 
  in function 
	(* Is this safe without unboxing? *)
(*    | Comparison (l, op, r, data) when constantp l && constantp r -> Some (Boolean ((assoc op ops) l r, data)) *)
    | Condition (Constant (Boolean true, _), t, _, _)  -> Some t
    | Condition (Constant (Boolean false, _), _, e, _) -> Some e
    | Concat (Nil _, c, _) 
    | Concat (c, Nil _, _) -> Some c
    | Concat (Constant(String l, _), Constant(String r, _), data) -> Some (Constant(String (l ^ r), data))
    | _ -> None 

(** Useful for printing the program at specific points of the
    optimisation pipeline.*)
let print_expression msg expr = 
  Debug.if_set show_optimisation
    (fun () -> msg ^ string_of_expression expr);
  None

(** Useful for checking a specific definition at specific points of the
    optimisation pipeline.*)
let print_definition of_name ?msg:msg def = 
 (match def with
    | Define (name, value, locn, _) when name = of_name 
        -> Debug.if_set show_optimisation
        (fun () -> fromOption "" msg ^ string_of_definition def)
    | _ -> ());
  None


let rewriters env = [
  uniquify_names; 
  RewriteSyntax.bottomup no_project_erase;  
  RewriteSyntax.bottomup renaming;
  RewriteSyntax.bottomup unused_variables;
  if Settings.get_value reduce_recs then
    RewriteSyntax.topdown reduce_recursion
  else RewriteSyntax.never;
  RewriteSyntax.topdown simplify_regex;
  RewriteSyntax.bottomup remove_unpacks;
  RewriteSyntax.topdown sql_aslist;
  RewriteSyntax.loop (RewriteSyntax.bottomup lift_lets);
  RewriteSyntax.topdown (sql_sort);
  RewriteSyntax.loop (RewriteSyntax.topdown sql_joins);
  RewriteSyntax.bottomup sql_selections;
  (*   RewriteSyntax.bottomup unused_variables; *)
  RewriteSyntax.bottomup sql_projections;
  (*   inference_rw env; *)
  RewriteSyntax.bottomup fold_constant;
  RewriteSyntax.topdown remove_trivial_extensions;
  RewriteSyntax.topdown (RewriteSyntax.both simplify_takedrop push_takedrop);
]

let run_optimisers : (Typeenv.environment * Typeenv.alias_environment) -> RewriteSyntax.rewriter
  = RewriteSyntax.all -<- rewriters

let optimise' env expr =
  if Settings.get_value optimising then 
    let expr' = run_optimisers env expr in
    let _ =
        match expr' with
            None -> Debug.if_set show_optimisation (fun () -> "Optimization had no effect")
          | Some expr' -> Debug.if_set show_optimisation
              (fun () ->
                 (if (Settings.get_value show_opt_verbose) then 
                    "Before optimization : " ^ 
                      Show_stripped_expression.show (strip_data expr) 
                  else "") ^ 
		   "\nAfter optimization  : " ^
                   Show_stripped_expression.show (strip_data expr'))
    in
      expr'
  else
    None
    
let optimise env expr = fromOption expr (optimise' env expr)

let optimise_program (omega, (Program (defs, body))) =
  let env = Typeenv.extract_typing_environment omega in
    Program (
      List.map (rewrite_def (optimise' env)) defs,
      optimise env body)
