#include <string.h>
#include "postgres.h"
#include "fmgr.h"
#include "libpq/pqformat.h"		/* needed for send/recv functions */

/* This exists in the current version PostgreSQL */
#define SET_VARSIZE(_PTR, _SIZE)  (((varattrib *)(_PTR))->va_header = _SIZE)

PG_MODULE_MAGIC;

/* Links labels as a PostgreSQL type */
typedef struct Label {
    int32 vl_len_;
    int32 num_args;
    int32 name_len;
    char name[1];
} Label;

Datum label_in(PG_FUNCTION_ARGS);
Datum label_out(PG_FUNCTION_ARGS);
Datum label_abs_eq(PG_FUNCTION_ARGS);
Datum label_abs_neq(PG_FUNCTION_ARGS);
Datum label_get_arg(PG_FUNCTION_ARGS);
Datum label_num_args(PG_FUNCTION_ARGS);
Datum label_get_name(PG_FUNCTION_ARGS);

static Label * parse_label(char*, int, int);
static char * label_to_string(Label*);
static bool label_match(Label*, Label*);
static Label * get_label_arg_internal(Label*, int);

static Label * get_label_arg_internal(Label *l, int n)
{
    int i;

    Label *p = (Label*)(l->name + l->name_len);;

    for (i = 0; i < l->num_args; ++i) {
        if (i == n) return p;
        p = (Label*)((char*)p + VARSIZE(p));
    }
    return NULL;
}


static Label * parse_label(char *p, int s, int e)
{
    int i;
    int pcount = 0;
    int len, old_size;
    
    Label *out, *l2, *q;

    /* -4: pretend last byte (rounded to word) doesn't exist */
    out = (Label *)palloc(sizeof(Label)-4); 
    SET_VARSIZE(out, sizeof(Label)-4);
    out->name_len = 0;
    out->num_args = 0;

    for (i = s; i < e; ++i) {
        switch (p[i]) {
        case '(': 
            if (pcount == 0) {
                len = i - s;    /* string length */
                /* parse name ws..i*/
                out->name_len = len+1 + ((len+1) % 4);
                SET_VARSIZE(out, VARSIZE(out) + out->name_len);
                out = (Label*)repalloc(out, VARSIZE(out));
                memcpy(out->name, p+s, len);
                out->name[len] = '\0';
                
                s = i+1;
            }
            pcount++;
            break;
        case ')':
            /* if pcount = 0, this is an error */
            pcount--;
            if (pcount > 0) break;
        case ',':
            if (pcount > 1 ) break;
            if (i-s >= 0) {  /* if we have parsed anything */
                l2 = parse_label(p, s, i);
                
                old_size = VARSIZE(out);
                SET_VARSIZE(out, VARSIZE(out) + VARSIZE(l2));
                out = (Label*)repalloc(out, VARSIZE(out));
                q = (Label*)(((char *)out) + old_size);
                (out->num_args)++;
                memcpy(q, l2, VARSIZE(l2));
                pfree(l2);
            }
            s = i+1;
            break;
        }
    }        

    if (out->name_len == 0) {
        len = e - s;    /* string length */
        out->name_len = len+1 + ((len+1) % 4);
        SET_VARSIZE(out, VARSIZE(out) + out->name_len);
        out = (Label*)repalloc(out, VARSIZE(out));
        memcpy(out->name, p+s, len);
        out->name[len] = '\0';
    }

    return out;
}

static char * label_to_string(Label *label) {
    Label *l2;
    char *result;
    char *p, *tmp;
    int i;
    
    result = (char*) palloc(strlen(label->name)+1+1); /* 1 for ( */
    p = stpcpy(result, label->name);
    
    if (label->num_args > 0) {
        p = stpcpy(p, "(");
        for (i = 0; i < label->num_args; ++i) {
            l2 = get_label_arg_internal(label, i);
            tmp = label_to_string(l2);

            result =            /* 2 for , and ) */
                (char*)repalloc(result, strlen(result) + strlen(tmp) + 1 + 2);
            p = result + strlen(result);
            
            if (i > 0)
                p = stpcpy(p, ",");
            p = stpcpy(p, tmp);
            pfree(tmp);
        }
        p = stpcpy(p, ")");
    }

    return(result);
}

static bool label_match(Label *l1, Label *l2)
{
    Label *c1, *c2;
    int i;
    
    const char* WILDCARD = "_";

    if (strncmp(l1->name, WILDCARD, 1) == 0 ||
        strncmp(l2->name, WILDCARD, 1) == 0)
        return true;            /* wildcards always match */
    else if (strcmp(l1->name, l2->name) == 0 &&
             l1->num_args == l2->num_args) {
        for (i = 0; i < l1->num_args; ++i) {
            c1 = get_label_arg_internal(l1, i);
            c2 = get_label_arg_internal(l2, i);
            if (!label_match(c1, c2))
                /* if any subexpression fail, match fails */
                return false;
        }
        return true;
    }
    return false;               /* default is false */
}

PG_FUNCTION_INFO_V1(label_in);

Datum
label_in(PG_FUNCTION_ARGS)
{
    char  *str = PG_GETARG_CSTRING(0);
    Label *result;

    result = parse_label(str, 0, strlen(str)+1);

    PG_RETURN_POINTER(result);
}    
    
PG_FUNCTION_INFO_V1(label_out);

Datum
label_out(PG_FUNCTION_ARGS)
{
    Label *label = (Label *) PG_GETARG_POINTER(0);
    char *result;
    
    result = label_to_string(label);
    PG_RETURN_POINTER(result);
}

                            
PG_FUNCTION_INFO_V1(label_abs_neq);

Datum
label_abs_neq(PG_FUNCTION_ARGS)
{
    Label *l1 = (Label *) PG_GETARG_POINTER(0);
    Label *l2 = (Label *) PG_GETARG_POINTER(1);
    
	PG_RETURN_BOOL(strcmp(l1->name, l2->name) != 0);
}

PG_FUNCTION_INFO_V1(label_get_arg);

Datum 
label_get_arg(PG_FUNCTION_ARGS)
{
    Label *result, *p;
    Label *l = (Label *) PG_GETARG_POINTER(0);
    int n = PG_GETARG_INT32(1);

    p = get_label_arg_internal(l, n);
    result = (Label*)palloc(VARSIZE(p));
    memcpy(result, p, VARSIZE(p));

    PG_RETURN_POINTER(result);
}

PG_FUNCTION_INFO_V1(label_get_name);

Datum
label_get_name(PG_FUNCTION_ARGS)
{
    Label *l = (Label *) PG_GETARG_POINTER(0);
    text *result;
    size_t result_size;
    char *name;
    
    name = l->name;

    if (strlen(name) == 0)
        name = "()";
    
    result_size = VARHDRSZ + strlen(name);
    result = (text*) palloc(result_size);
    SET_VARSIZE(result, result_size);
    memcpy(VARDATA(result), name, VARSIZE(result) - VARHDRSZ);

    PG_RETURN_TEXT_P(result);
}


PG_FUNCTION_INFO_V1(label_num_args);

Datum
label_num_args(PG_FUNCTION_ARGS)
{
    Label *l = (Label *) PG_GETARG_POINTER(0);
    int n = l->num_args;

    PG_RETURN_INT32(n);
}


PG_FUNCTION_INFO_V1(label_abs_eq);

Datum
label_abs_eq(PG_FUNCTION_ARGS)
{
    Label *l1 = (Label *) PG_GETARG_POINTER(0);
    Label *l2 = (Label *) PG_GETARG_POINTER(1);

    PG_RETURN_BOOL(label_match(l1,l2));
}

