(*pp deriving *)

open Expression
open Expression.Expr
  
module Sql : Sql.S with module Query = Query

type position = Expression.position
type location = [`Client | `Server | `Native | `Policy | `Database | `Unknown]
    deriving (Eq, Typeable, Show, Pickle, Shelve)

type 'a definition' =
  | Define of (string * 'a expression' * location * 'a)
  | Alias of (string * (int * FableType.kind) list * 'a Types.datatype' * 'a)
  | Alien of (string * string * 'a Types.assumption' * 'a)
  | Module of (string * 'a)

type 'a program' = Program of ('a definition' list * 'a expression')

val show_pos : position -> string

type typed_data = Types.datatype annotated_data 
type expression = FableType.typed_expression 
    deriving (Eq, Show)

type table_constr = typed_data Expression.Expr.table_constr

type untyped_expression = FableType.untyped_expression
type stripped_expression = unit expression' (* XXX deriving (Show) *)

module Show_stripped_expression : Show with type a = stripped_expression

type definition = typed_data definition' 
    deriving (Show) 

type untyped_definition = untyped_data definition'
type stripped_definition = unit definition'

type program = typed_data program' 
type untyped_program = untyped_data program' (* XXX deriving (Show) *)
type stripped_program = unit program' (* XXX deriving (Show) *)

exception ASTSyntaxError of position * string

val unit_expression : 'a -> 'a expression'
val is_unit : 'a expression' -> bool

val list_expr : 'a -> 'a expression' list -> 'a expression'
  
val tuple_expr : 'a -> 'a expression' list -> 'a expression'

val is_value : 'a expression' -> bool

val defined_names : 'a definition' list -> string list

val string_of_expression : 'a expression' -> string
val labelled_string_of_expression : expression -> string

val string_of_definition : 'a definition' -> string
val labelled_string_of_definition : definition -> string

val string_of_program : 'a program' -> string
val labelled_string_of_program : program -> string

val stringlit_value : 'a expression' -> string

val freevars : 'a expression' -> Utility.StringSet.t
val freevars_list : 'a expression' -> string list
val freevars_def : 'a definition' -> Utility.StringSet.t
val freevars_program : 'a program' -> Utility.StringSet.t

val free_bound_type_vars : Types.typed_data expression' -> Types.TypeVarSet.t
val free_bound_type_vars_def : Types.typed_data definition' -> Types.TypeVarSet.t
val free_bound_type_vars_program : Types.typed_data program' -> Types.TypeVarSet.t

(** {0 Variable-substitution functions} 
    TBD: gen'ize these for typed & other exprs as well *)
val subst_free : string -> untyped_expression -> untyped_expression -> untyped_expression
val rename_free : string -> string -> untyped_expression -> untyped_expression
val subst_fast : string -> expression -> expression -> expression
val rename_fast : string -> string -> expression -> expression

val subst_fast_def : string -> expression -> definition -> definition

val reduce_expression :
  (('a expression' -> 'b) -> 'a expression' -> 'b) ->
  ('a expression' * 'b list -> 'b) -> 'a expression' -> 'b
val reduce_definition :
  (('a expression' -> 'b) -> 'a expression' -> 'b) ->
  ('a expression' * 'b list -> 'b) ->
  ('a definition' * 'b list -> 'b) ->
  'a definition' -> 'b
val reduce_program :
  (('a expression' -> 'b) -> 'a expression' -> 'b) ->
  ('a expression' * 'b list -> 'b) ->
  ('a definition' * 'b list -> 'b) ->
  ('b list * 'b -> 'b) ->
  'a program' -> 'b

val set_subnodes : 'a expression' -> 'a expression' list -> 'a expression'
  
val expression_data : 'a expression' -> 'a
val strip_data : 'a expression' -> stripped_expression
val node_datatype : expression -> Types.datatype
val def_datatype : definition -> Types.datatype

val set_node_datatype : (expression * Types.datatype) -> expression

type data = [untyped_data | typed_data]

val data_position : [<data] -> position
val position : [<data] expression' -> position

val erase : expression -> untyped_expression
val labelize : program -> program
  
val dummy_position : position
val no_expr_data : typed_data

val is_closed : expression -> bool
val is_closed_wrt : expression -> Utility.StringSet.t -> bool

module RewriteUntypedExpression : Rewrite.Rewrite with type t = untyped_expression
module RewriteSyntax : Rewrite.Rewrite with type t = expression

val transform_def :
  ('a expression' -> 'a expression') -> 'a definition' -> 'a definition'
val transform_program :
  ('a expression' -> 'a expression') -> 'a program' -> 'a program'

val rewrite_def :
  ('a expression' -> 'a expression' option) -> 'a definition' -> 'a definition'
val rewrite_program :
  ('a expression' -> 'a expression' option) -> 'a program' -> 'a program'

  
